/* $Header: /home/jcb/newmj/RCS/gui-dial.c,v 11.21 2002/02/04 22:44:28 jcb Rel $
 * gui-dial.c
 * dialog box functions.
 */
/****************** COPYRIGHT STATEMENT **********************
 * This file is Copyright (c) 2000 by J. C. Bradfield.       *
 * Distribution and use is governed by the LICENCE file that *
 * accompanies this file.                                    *
 * The moral rights of the author are asserted.              *
 *                                                           *
 ***************** DISCLAIMER OF WARRANTY ********************
 * This code is not warranted fit for any purpose. See the   *
 * LICENCE file for further information.                     *
 *                                                           *
 *************************************************************/

#include "gui.h"

static const char rcs_id[] = "$Header: /home/jcb/newmj/RCS/gui-dial.c,v 11.21 2002/02/04 22:44:28 jcb Rel $";

static void continue_callback(GtkWidget *w, gpointer data);
static void turn_callback(GtkWidget *w, gpointer data);


/* Used sordidly and internally */
static GtkRequisition discard_req = { 0, 0};
/* Why an array? So I can pass pointers around */
DiscardDialog discard_dialog[1];

/* dialog box for specifying chows */
GtkWidget *chow_dialog;
/* Stores the three TileSetBoxes */
static TileSetBox chowtsbs[3];
/* and the three buttons */
static GtkWidget *chowbuttons[3];

/* dialog box for declaring specials */
GtkWidget *ds_dialog;

/* dialog box for continuing with next hand */
GtkWidget *continue_dialog;

/* dialog for opening connection */
GtkWidget *open_dialog;
GtkWidget *openmenuentry, *newgamemenuentry, *resumegamemenuentry, 
  *savemenuentry, *saveasmenuentry, *closemenuentry, *gameoptionsmenuentry;

/* dialog box for action when it's our turn.
   Actions: Discard  Kong  Add to Pung  Mah Jong
*/
GtkWidget *turn_dialog;
GtkWidget *turn_dialog_label; /* used to display number of tiles left */

/* dialog box for closed sets when scoring.
   Actions: Eyes  Chow  Pung  Done
*/
GtkWidget *scoring_dialog;

/* window for game status display */
GtkWidget *status_window;

/* window for "about" information */
GtkWidget *about_window;
/* window to nag for donations */
GtkWidget *nag_window;

/* an array of text widgets for displaying scores etc.
   Element 4 is for settlements.
   The others are for each player: currently, I think
   these should be table relative.
*/
GtkWidget *scoring_notebook;
GtkWidget *textpages[5];
GtkWidget *textlabels[5]; /* labels for the pages */
GtkWidget *textwindow; /* and the window for it */

/* The window for messages, and the display text widget */
GtkWidget *messagewindow, *messagetext;

/* The Save As.. dialog */
GtkWidget *save_window;
/* and its text entry widget */
GtkWidget *save_text;

/* The window for display options */
GtkWidget *display_option_dialog = NULL;

/* the window for updating the game options */
GtkWidget *game_option_dialog = NULL;
GtkWidget *game_option_panel = NULL;
/* and some of its buttons */
GtkWidget *game_option_apply_button = NULL;
GtkWidget *game_option_prefs_button = NULL;

/* and a very similar one for option preferences */
GtkWidget *game_prefs_dialog = NULL;
GtkWidget *game_prefs_panel = NULL;

/* time at which progress bar was started */
static struct timeval pstart;
static int pinterval = 25; /* timeout interval in ms */
static int pbar_timeout_instance = 0; /* track dead timeouts */
static GtkWidget *pbar;

/* timeout handler for the dialog progress bar */
static int pbar_timeout(gpointer instance) {
  int timeleft;
  struct timeval timenow;

  if ( pbar_timeout_instance != (int) instance ) return FALSE; /* dead timeout */
  if ( ! GTK_WIDGET_VISIBLE(discard_dialog->widget) ) return FALSE;
  gettimeofday(&timenow,NULL);
  timeleft = ptimeout-(1000*(timenow.tv_sec-pstart.tv_sec)
	      +(timenow.tv_usec-pstart.tv_usec)/1000);
  if ( timeleft <= 0 ) {
    /* we should not hide the claim dialog: the timeout is really
       controlled by the server, not us */
    return FALSE;
  }
  gtk_progress_bar_update(GTK_PROGRESS_BAR(pbar),1.0-(timeleft+0.0)/(ptimeout+0.0));
  return TRUE;
}

/* popup the discard dialog. Arguments:
   Tile, player whence it came (as an ori),
   mode = 0 (normal), 1 (claiming tile for mah jong), 2 (claiming from kong) */
void discard_dialog_popup(Tile t, int ori, int mode) {
  gint x,y,w,h;
  int i;
  char buf[128];
  static Tile lastt; static int lastori, lastmode;
  static int positioned = 0;

  /* So that we don't work if it's already popped up: */
  if ( GTK_WIDGET_VISIBLE(discard_dialog->widget)
       && t == lastt && lastori == ori && lastmode == mode ) return;
  lastt = t; lastori = ori; lastmode = mode;

  if ( mode != discard_dialog->mode ) {
    discard_dialog->mode = mode;
    if ( mode == 0 ) {
      gtk_widget_show(discard_dialog->noclaim);
      gtk_widget_hide(discard_dialog->eyes);
      gtk_widget_show(discard_dialog->chow);
      gtk_widget_show(discard_dialog->pung);
      gtk_widget_hide(discard_dialog->special);
      gtk_widget_show(discard_dialog->kong);
      gtk_widget_show(discard_dialog->mahjong);
      gtk_widget_hide(discard_dialog->robkong);
    } else if ( mode == 1 ) {
      gtk_widget_hide(discard_dialog->noclaim);
      gtk_widget_show(discard_dialog->eyes);
      gtk_widget_show(discard_dialog->chow);
      gtk_widget_show(discard_dialog->pung);
      gtk_widget_show(discard_dialog->special);
      gtk_widget_hide(discard_dialog->kong);
      gtk_widget_hide(discard_dialog->mahjong);
      gtk_widget_hide(discard_dialog->robkong);
    } else {
      gtk_widget_show(discard_dialog->noclaim);
      gtk_widget_hide(discard_dialog->eyes);
      gtk_widget_hide(discard_dialog->chow);
      gtk_widget_hide(discard_dialog->pung);
      gtk_widget_hide(discard_dialog->special);
      gtk_widget_hide(discard_dialog->kong);
      gtk_widget_hide(discard_dialog->mahjong);
      gtk_widget_show(discard_dialog->robkong);
    }
  }
  if ( mode == 0 ) gtk_widget_grab_focus(discard_dialog->noclaim);

  /* center the dialog on the main window */
  /* Oh. I'd like to tell the window manager not to shift it
     when it puts the border round it. I wonder how I do that?
  */
  /* Note that the size has been fixed in the init routine.
     The reason we do this statically is so that we can
     force the bloody thing to have the right size: we want to
     hide buttons before showing it, but then it would have a
     different size on being shown.
     Also, we want it to have a fixed size so we can put
     the turn dialog to coincide with this one's bottom left.
  */
  switch ( dialogs_position ) {
  case DialogsCentral:
  case DialogsUnspecified:
    /* dialog is child of discard area */
    w = discard_area_alloc.width;
    h = discard_area_alloc.height;
    x = y = 0;
    gtk_widget_set_uposition(discard_dialog->widget,
			     x + w/2 - discard_req.width/2,
			     y + h/2 - discard_req.height/2);
    break;
  case DialogsPopup:
    /* only do this the first time; we should let the user position them */
    if ( ! positioned++ ) {
      gdk_window_get_size(topwindow->window,&w,&h);
      gdk_window_get_deskrelative_origin(topwindow->window,&x,&y);
      gtk_widget_set_uposition(discard_dialog->widget,
			       x + w/2 - discard_req.width/2,
			       y + h/2 - discard_req.height/2);
    }
    break;
  case DialogsBelow:
    ;
  }

  /* set the appropriate tile */
  for ( i=1 ; i < 4 ; i++ ) {
    if ( i == ori ) {
      button_set_tile(discard_dialog->tiles[i],t,i);
      gtk_widget_show(discard_dialog->tiles[i]);
    } else {
      gtk_widget_hide(discard_dialog->tiles[i]);
    }
  }
  gtk_widget_hide(discard_dialog->tilename);
  if ( mode == 1 ) {
    gtk_label_set_text(GTK_LABEL(discard_dialog->tilename),
		     "Claim discard for:");
  } else {
    /* if not showing wall, say how many tiles left */
    if ( ! showwall ) {
      if ( ori == 1 ) sprintf(buf,"(%d tiles left) %s",
			      the_game->wall.live_end-the_game->wall.live_used,
			      tile_name(the_game->tile));
      else sprintf(buf,"%s (%d tiles left)",
		   tile_name(the_game->tile),
		   the_game->wall.live_end-the_game->wall.live_used);
    } else { 
      strcpy(buf,tile_name(the_game->tile));
    }
    gtk_label_set_text(GTK_LABEL(discard_dialog->tilename),buf);
  }
  if ( mode == 0 ) {
    static const gfloat xal[] = { 0.5,1.0,0.5,0.0 }; 
    gtk_misc_set_alignment(GTK_MISC(discard_dialog->tilename),
			  xal[ori],0.5);
  } else {
    gtk_misc_set_alignment(GTK_MISC(discard_dialog->tilename),
			  0.5,0.5);
  }
  gtk_widget_show(discard_dialog->widget);
  gtk_widget_show(discard_dialog->tilename);
  /* and start the progress bar timeout if appropriate */
  if ( the_game->state != MahJonging && ptimeout > 0 ) {
    gtk_widget_show(pbar);
    gettimeofday(&pstart,NULL);
    /* we may as well calculate an appropriate value of pbar_timeout
       each time... we want it to update every half pixel, or 40 times
       a second, whichever is slower */
    if ( pbar->allocation.width > 1 ) {
      /* in case it isn't realized yet */
      pinterval = ptimeout/(2*pbar->allocation.width);
    }
    if ( pinterval < 25 ) pinterval = 25;
    gtk_timeout_add(pinterval,pbar_timeout,(gpointer) ++pbar_timeout_instance);
  } else {
    gtk_widget_hide(pbar);
  }
}

/* an accelerator group  for the discard dialog */
static GtkAccelGroup *discard_accel;

/* this little function is used to install or remove the
   discard accelerators */
static void add_or_remove_discard_accels(GtkWidget *w UNUSED, gpointer data)
{
  if ( data )
    gtk_window_add_accel_group(GTK_WINDOW(topwindow),discard_accel);
  else
    gtk_window_remove_accel_group(GTK_WINDOW(topwindow),discard_accel);
}

/* initialize it */
/* Structure:
   If the dialog is in the middle:
      lefttile opptile righttile
           tile name
         progress bar
      Pass/Draw Chow Pung Kong MahJong
   otherwise: 
      tilename   progress bar
         buttons
*/
void discard_dialog_init(void) {
  GtkWidget *box, *tilebox, *left, *opp, *right, *lbl,
      *butbox, *but, *pixm;
  DiscardDialog *dd = &discard_dialog[0];

  if ( dd->widget ) {
    gtk_widget_destroy(dd->widget);
    dd->widget = NULL;
  }

  switch ( dialogs_position ) {
  case DialogsCentral:
  case DialogsUnspecified:
    /* event box so there's a window to have background */
    dd->widget = gtk_event_box_new();
    gtk_fixed_put(GTK_FIXED(discard_area),dd->widget,0,0);
    /* it'll be moved later */
    break;
  case DialogsBelow:
    dd->widget = gtk_event_box_new();
    gtk_box_pack_start(GTK_BOX(dialoglowerbox),dd->widget,1,0,0);
    /* show it, so that the top window includes it when first mapped */
    gtk_widget_show(dd->widget);
    break;
  case DialogsPopup:
    dd->widget = gtk_window_new(GTK_WINDOW_DIALOG);
    gtk_signal_connect (GTK_OBJECT (dd->widget), "delete_event",
                               GTK_SIGNAL_FUNC (gtk_widget_hide), NULL);
  }

  box = gtk_vbox_new(0,dialog_vert_spacing);
  gtk_widget_show(box);
  gtk_container_add(GTK_CONTAINER(dd->widget),box);
  dd->mode = -1; /* so that personality will be set */
  gtk_container_set_border_width(GTK_CONTAINER(box),
				 dialog_border_width);

  tilebox = gtk_hbox_new(0,0);
  if ( dialogs_position != DialogsBelow ) gtk_widget_show(tilebox);
  
  left = gtk_button_new();
  GTK_WIDGET_UNSET_FLAGS(left,GTK_CAN_FOCUS);
  gtk_widget_show(left);
  pixm = gtk_pixmap_new(tilepixmaps[3][HiddenTile],NULL);
  gtk_widget_show(pixm);
  gtk_container_add(GTK_CONTAINER(left),pixm);
  opp = gtk_button_new();
  GTK_WIDGET_UNSET_FLAGS(opp,GTK_CAN_FOCUS);
  gtk_widget_show(opp);
  pixm = gtk_pixmap_new(tilepixmaps[2][HiddenTile],NULL);
  gtk_widget_show(pixm);
  gtk_container_add(GTK_CONTAINER(opp),pixm);
  right = gtk_button_new();
  GTK_WIDGET_UNSET_FLAGS(right,GTK_CAN_FOCUS);
  gtk_widget_show(right);
  pixm = gtk_pixmap_new(tilepixmaps[1][HiddenTile],NULL);
  gtk_widget_show(pixm);
  gtk_container_add(GTK_CONTAINER(right),pixm);
  gtk_box_pack_start(GTK_BOX(tilebox),left,0,0,0);
  gtk_box_pack_start(GTK_BOX(tilebox),opp,1,0,0);
  gtk_box_pack_end(GTK_BOX(tilebox),right,0,0,0);
  dd->tiles[1] = right;
  dd->tiles[2] = opp;
  dd->tiles[3] = left;
  
  lbl = gtk_label_new("name of tile");
  gtk_widget_show(lbl);
  dd->tilename = lbl;

  butbox = gtk_hbox_new(1,dialog_button_spacing); /* homogeneous, spaced */
  gtk_widget_show(butbox);
  
  /* accelerators for discard dialog actions */
  discard_accel = gtk_accel_group_new();

  but = gtk_button_new_with_label("No claim");
  gtk_widget_show(but);
  gtk_box_pack_start(GTK_BOX(butbox),but,1,1,0);
  gtk_signal_connect(GTK_OBJECT(but),"clicked",
		     disc_callback,(gpointer)NoClaim);
  dd->noclaim = but;
  gtk_accel_group_add(discard_accel,GDK_n,0,0,GTK_OBJECT(but),"clicked");
  gtk_label_set_pattern(GTK_LABEL(GTK_BIN(but)->child),"_");

  but = gtk_button_new_with_label("Eyes");
  GTK_WIDGET_UNSET_FLAGS(but,GTK_CAN_FOCUS);
  /* not shown in normal state */
  gtk_box_pack_start(GTK_BOX(butbox),but,1,1,0);
  gtk_signal_connect(GTK_OBJECT(but),"clicked"
		     ,disc_callback,(gpointer)PairClaim);
  dd->eyes = but;
  gtk_accel_group_add(discard_accel,GDK_e,0,0,GTK_OBJECT(but),"clicked");
  gtk_label_set_pattern(GTK_LABEL(GTK_BIN(but)->child),"_");

  but = gtk_button_new_with_label("Chow");
  GTK_WIDGET_UNSET_FLAGS(but,GTK_CAN_FOCUS);
  gtk_widget_show(but);
  gtk_box_pack_start(GTK_BOX(butbox),but,1,1,0);
  gtk_signal_connect(GTK_OBJECT(but),"clicked"
		     ,disc_callback,(gpointer)ChowClaim);
  dd->chow = but;
  gtk_accel_group_add(discard_accel,GDK_c,0,0,GTK_OBJECT(but),"clicked");
  gtk_label_set_pattern(GTK_LABEL(GTK_BIN(but)->child),"_");

  but = gtk_button_new_with_label("Pung");
  GTK_WIDGET_UNSET_FLAGS(but,GTK_CAN_FOCUS);
  gtk_widget_show(but);
  gtk_box_pack_start(GTK_BOX(butbox),but,1,1,0);
  gtk_signal_connect(GTK_OBJECT(but),"clicked",
		     disc_callback,(gpointer)PungClaim);
  dd->pung = but;
  gtk_accel_group_add(discard_accel,GDK_p,0,0,GTK_OBJECT(but),"clicked");
  gtk_label_set_pattern(GTK_LABEL(GTK_BIN(but)->child),"_");

  but = gtk_button_new_with_label("Special Hand");
  GTK_WIDGET_UNSET_FLAGS(but,GTK_CAN_FOCUS);
  /* gtk_widget_show(but); */ /* don't show this; uses other space */
  gtk_box_pack_start(GTK_BOX(butbox),but,1,1,0);
  gtk_signal_connect(GTK_OBJECT(but),"clicked",
		     disc_callback,(gpointer)SpecialSetClaim);
  dd->special = but;
  gtk_accel_group_add(discard_accel,GDK_s,0,0,GTK_OBJECT(but),"clicked");
  gtk_label_set_pattern(GTK_LABEL(GTK_BIN(but)->child),"_");

  but = gtk_button_new_with_label("Kong");
  GTK_WIDGET_UNSET_FLAGS(but,GTK_CAN_FOCUS);
  gtk_widget_show(but);
  gtk_box_pack_start(GTK_BOX(butbox),but,1,1,0);
  gtk_signal_connect(GTK_OBJECT(but),"clicked",
		     disc_callback,(gpointer)KongClaim);
  dd->kong = but;
  gtk_accel_group_add(discard_accel,GDK_k,0,0,GTK_OBJECT(but),"clicked");
  gtk_label_set_pattern(GTK_LABEL(GTK_BIN(but)->child),"_");

  but = gtk_button_new_with_label("Mah Jong");
  GTK_WIDGET_UNSET_FLAGS(but,GTK_CAN_FOCUS);
  gtk_widget_show(but);
  gtk_box_pack_start(GTK_BOX(butbox),but,1,1,0);
  gtk_signal_connect(GTK_OBJECT(but),"clicked",
		     disc_callback,(gpointer)MahJongClaim);
  dd->mahjong = but;
  gtk_accel_group_add(discard_accel,GDK_m,0,0,GTK_OBJECT(but),"clicked");
  gtk_label_set_pattern(GTK_LABEL(GTK_BIN(but)->child),"_");

  but = gtk_button_new_with_label("Rob the Kong - Mah Jong!");
  GTK_WIDGET_UNSET_FLAGS(but,GTK_CAN_FOCUS);
  /* gtk_widget_show(but); */ /* don't show this; it uses the space of others */
  gtk_box_pack_start(GTK_BOX(butbox),but,1,1,0);
  gtk_signal_connect(GTK_OBJECT(but),"clicked",
		     disc_callback,(gpointer)MahJongClaim);
  dd->robkong = but;
  gtk_accel_group_add(discard_accel,GDK_r,0,0,GTK_OBJECT(but),"clicked");
  gtk_label_set_pattern(GTK_LABEL(GTK_BIN(but)->child),"_");

  pbar = gtk_progress_bar_new();
  gtk_widget_show(pbar);

  /* These are packed in reverse order so they float to the bottom */
  gtk_box_pack_end(GTK_BOX(box),butbox,0,0,0);
  gtk_box_pack_end(GTK_BOX(box),lbl,0,0,0);
  gtk_box_pack_end(GTK_BOX(box),pbar,0,0,0);
  gtk_box_pack_end(GTK_BOX(box),tilebox,0,0,0);
  
  /* OK, now ask its size: store the result, and keep it
     this size for ever more */
  gtk_widget_size_request(dd->widget,&discard_req);
  gtk_widget_set_usize(dd->widget,discard_req.width,discard_req.height);
  
  /* we have to ensure that the accelerators are only available when
     the widget is popped up */
  gtk_signal_connect(GTK_OBJECT(discard_dialog->widget),"hide",
    add_or_remove_discard_accels,(gpointer)0);
  gtk_signal_connect(GTK_OBJECT(discard_dialog->widget),"show",
    add_or_remove_discard_accels,(gpointer)1);

}

static GtkWidget *turn_dialog_discard_button;
static GtkWidget *turn_dialog_calling_button;

void turn_dialog_popup(void) {
  /* only original call is allowed, so hide the calling
     button after first discard */
  if ( pflag(our_player,NoDiscard) ) 
    gtk_widget_show(turn_dialog_calling_button);
  else 
    gtk_widget_hide(turn_dialog_calling_button);
  /* if not showing wall, put tiles left in label */
  if ( ! showwall ) {
    char buf[128];
    sprintf(buf,"(%d tiles left)  Select tile and:",
	    the_game->wall.live_end-the_game->wall.live_used);
    gtk_label_set_text(GTK_LABEL(turn_dialog_label),buf);
  } else {
    gtk_label_set_text(GTK_LABEL(turn_dialog_label),"Select tile and:");
  }
  dialog_popup(turn_dialog,DPOnDiscardOnce);
  gtk_widget_grab_focus(turn_dialog_discard_button);
}

/* callback when we toggle one of our tiles.
   data is our index number.
   If data also has bit 7 set, force the tile active.
   If called with NULL widget, clear selection 
*/

void conc_callback(GtkWidget *w, gpointer data) {
  int active;
  int i;
  int force=0, index=-1;
  static GtkWidget *selected = NULL; /* for radiogroup functionality */

  active = (w && gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w)));
  if ( active ) index = ((int)data)& 127;
  force = (w && (((int)data) & 128));


  if ( w && just_doubleclicked == w) force = 1;

  /* make sure all other tiles are unselected, if we're active,
     or if we're called to clear: we don't just rely
     on the selected variable, since under some circumstances we
     can end up with two tiles active, by accident as it were */
  /* FIXME: this relies on induced callbacks being executed synchronously */
  if ( active || w == NULL) {
    for ( i = 0; i<14; i++ ) {
      if ( pdisps[0].conc[i] && pdisps[0].conc[i] != w) {
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pdisps[0].conc[i]),
				     FALSE);
      }
    }
  }
  selected_button = index;
  selected = NULL;
  if ( active ) selected = w;

  if ( w == NULL ) return;

  if ( force ) {
    /* if it's not active, set it active: we'll then
       be invoked normally, so we just return now. */
    if ( ! active ) {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w),TRUE);
      return;
    }
  }

  /* if we were double clicked, invoke the turn callback directly */
  if ( w && just_doubleclicked == w) {
    just_doubleclicked = 0;
    turn_callback(w,(gpointer)PMsgDiscard);
  }
}

/* This detects doubleclicks on the concealed buttons */
gint doubleclicked(GtkWidget *w, GdkEventButton *eb,gpointer data UNUSED) {
  if ( eb->type != GDK_2BUTTON_PRESS ) return FALSE;
  /* This is disgusting. We set a global doubleclicked flag,
     which is noticed by the toggle callback */
  just_doubleclicked = w;
  return FALSE;
}

/* callback attached to the buttons of the discarding dialog.
   They pass PMsgDiscard, PMsgDeclareClosedKong,
   PMsgAddToPung, or PMsgMahJong.
   Passed PMsgDiscard + 1000000 to declare calling.
   Also invoked by the declaring special callback:
    with DeclareSpecial to declare a special, Kong if appropriate,
    and NoClaim to indicate the end of declaration.
   Also invoked by scoring dialog with appropriate values */

static void turn_callback(GtkWidget *w UNUSED, gpointer data) {
  PMsgUnion m;
  Tile selected_tile;

  m.type = (PlayerMsgType)data;

  if ( m.type == PMsgMahJong 
       || m.type == PMsgShowTiles
       || m.type == PMsgFormClosedSpecialSet ) {
    send_packet(&m);
    return;
  }

  if ( the_game->state == DeclaringSpecials
       && m.type == PMsgNoClaim ) {
    m.type = PMsgDeclareSpecial;
    conc_callback(NULL,NULL); /* clear the selection */
  }

  selected_tile = (selected_button < 0) ? HiddenTile : our_player->concealed[selected_button];

  /* in declaring specials, use this to finish */
  if ( selected_tile == HiddenTile ) {
    if ( the_game->state == DeclaringSpecials )
      m.type = PMsgDeclareSpecial;
    else {
      error_dialog_popup("No tile selected!");
      return;
    }
  }
  if ( is_special(selected_tile) )
    m.type = PMsgDeclareSpecial;

  switch ( m.type ) {
  case PMsgDeclareSpecial:
    m.declarespecial.tile = selected_tile;
    break;
  case PMsgDiscard:
    m.discard.tile = selected_tile;
    m.discard.calling = 0;
    break;
  case PMsgDiscard+1000000:
    m.type = PMsgDiscard;
    m.discard.tile = selected_tile;
    m.discard.calling = 1;
    break;
  case PMsgDeclareClosedKong:
    m.declareclosedkong.tile = selected_tile;
    break;
  case PMsgAddToPung:
    m.addtopung.tile = selected_tile;
    break;
  case PMsgFormClosedPair:
    m.formclosedpair.tile = selected_tile;
    break;
  case PMsgFormClosedChow:
    m.formclosedchow.tile = selected_tile;
    break;
  case PMsgFormClosedPung:
    m.formclosedpung.tile = selected_tile;
    break;
  default:
    warn("bad type in turn_callback");
    return;
  }
  send_packet(&m);
}
  
/* callback when one of the discard dialog buttons is clicked */
void disc_callback(GtkWidget *w UNUSED, gpointer data) {
  PMsgUnion m;

  switch ( (gint32) data) {
  case NoClaim:
    m.type = PMsgNoClaim;
    m.noclaim.discard = the_game->serial;
    break;
  case ChowClaim:
    m.type = PMsgChow;
    m.chow.discard = the_game->serial;
    m.chow.cpos = AnyPos; /* worry about it later */
    break;
  case PairClaim:
    m.type = PMsgPair;
    break;
  case SpecialSetClaim:
    m.type = PMsgSpecialSet;
    break;
  case PungClaim:
    m.type = PMsgPung;
    m.pung.discard = the_game->serial;
    break;
  case KongClaim:
    m.type = PMsgKong;
    m.kong.discard = the_game->serial;
    break;
  case MahJongClaim:
    m.type = PMsgMahJong;
    m.mahjong.discard = the_game->serial;
    break;
  default:
    warn("disc callback called with unexpected data");
    return;
  }

  /* if we're mahjonging, we mustn't send an anypos,
     so just call do_chow immediately */
  if ( the_game->state == MahJonging && m.type == PMsgChow ) {
    do_chow(NULL,(gpointer)AnyPos);
  } else {
    send_packet(&m);
  }
}

static GtkAccelGroup *chow_accel;

static void add_or_remove_chow_accels(GtkWidget *w UNUSED, gpointer data)
{
  if ( data )
    gtk_window_add_accel_group(GTK_WINDOW(topwindow),chow_accel);
  else
    gtk_window_remove_accel_group(GTK_WINDOW(topwindow),chow_accel);
}

  /* Now create the Chow dialog box:
     Structure: three boxes, each containing a tileset showing
     a possible chow. Below that, a label "which chow?".
     Below that, three buttons to select.
  */
void chow_dialog_init(void)  {
  GtkWidget *box,*u,*v; int i;

  if ( chow_dialog ) {
    gtk_widget_destroy(chow_dialog);
    chow_dialog = NULL;
  }

  switch ( dialogs_position ) {
  case DialogsCentral:
  case DialogsUnspecified:
    chow_dialog = gtk_event_box_new();
    gtk_fixed_put(GTK_FIXED(discard_area),chow_dialog,0,0);
    break;
  case DialogsPopup:
  case DialogsBelow:
    chow_dialog = gtk_window_new(GTK_WINDOW_DIALOG);
    gtk_signal_connect (GTK_OBJECT (chow_dialog), "delete_event",
                               GTK_SIGNAL_FUNC (gtk_widget_hide), NULL);
    /* This one is allowed to shrink, and should */
    gtk_window_set_policy(GTK_WINDOW(chow_dialog),1,1,1);
  }
  box = gtk_vbox_new(0,dialog_vert_spacing);
  gtk_container_set_border_width(GTK_CONTAINER(box),
				 dialog_border_width);
  gtk_widget_show(box);
  gtk_container_add(GTK_CONTAINER(chow_dialog),box);
  u = gtk_hbox_new(0,dialog_button_spacing);
  gtk_widget_show(u);
  gtk_box_pack_start(GTK_BOX(box),u,0,0,0);
  for (i=0;i<3;i++) {
    v = gtk_hbox_new(0,0);
    gtk_widget_show(v);
    gtk_box_pack_start(GTK_BOX(u),v,0,0,0);
    chowtsbs[i].widget = v;
    tilesetbox_init(&chowtsbs[i],0,do_chow,(gpointer)i);
  }
  u = gtk_label_new("Which chow?");
  gtk_widget_show(u);
  gtk_box_pack_start(GTK_BOX(box),u,0,0,0);
  u = gtk_hbox_new(1,dialog_button_spacing);
  gtk_widget_show(u);
  gtk_box_pack_start(GTK_BOX(box),u,0,0,0);

  chow_accel = gtk_accel_group_new();

  for (i=0;i<3;i++) {
    static int keys[] = { GDK_l, GDK_m, GDK_u };
    v = gtk_button_new_with_label(player_print_ChowPosition(i));
    GTK_WIDGET_UNSET_FLAGS(v,GTK_CAN_FOCUS);
    gtk_widget_show(v);
    gtk_box_pack_start(GTK_BOX(u),v,1,1,0);
    chowbuttons[i] = v;
    gtk_signal_connect(GTK_OBJECT(v),"clicked",
		       do_chow,(gpointer)i);
    gtk_accel_group_add(chow_accel,keys[i],0,0,GTK_OBJECT(v),"clicked");
    gtk_label_set_pattern(GTK_LABEL(GTK_BIN(v)->child),"_");
  }

  gtk_signal_connect(GTK_OBJECT(chow_dialog),"hide",
    add_or_remove_chow_accels,(gpointer)0);
  gtk_signal_connect(GTK_OBJECT(chow_dialog),"show",
    add_or_remove_chow_accels,(gpointer)1);

}

/* now create the declaring specials dialog.
   Contains a button for declare (special), kong, and done.
   If no flowers, just kongs.
*/
static GtkWidget *ds_dialog_declare_button;
static GtkWidget *ds_dialog_finish_button;

static GtkAccelGroup *ds_accel;

static void add_or_remove_ds_accels(GtkWidget *w UNUSED, gpointer data)
{
  if ( data )
    gtk_window_add_accel_group(GTK_WINDOW(topwindow),ds_accel);
  else
    gtk_window_remove_accel_group(GTK_WINDOW(topwindow),ds_accel);
}

void ds_dialog_init(void) {
  GtkWidget *box, *bbox, *w;

  if ( ds_dialog ) {
    gtk_widget_destroy(ds_dialog);
    ds_dialog = NULL;
  }

  switch ( dialogs_position ) {
  case DialogsCentral:
  case DialogsUnspecified:
    ds_dialog = gtk_event_box_new();
    gtk_fixed_put(GTK_FIXED(discard_area),ds_dialog,0,0);
    break;
  case DialogsBelow:
    ds_dialog = gtk_event_box_new();
    gtk_box_pack_start(GTK_BOX(dialoglowerbox),ds_dialog,1,0,0);
    break;
  case DialogsPopup:
    ds_dialog = gtk_window_new(GTK_WINDOW_DIALOG);
    gtk_signal_connect (GTK_OBJECT (ds_dialog), "delete_event",
                               GTK_SIGNAL_FUNC (gtk_widget_hide), NULL);
  }
  
  box = gtk_vbox_new(0,dialog_vert_spacing);
  gtk_container_set_border_width(GTK_CONTAINER(box),
				 dialog_border_width);
  gtk_widget_show(box);
  gtk_container_add(GTK_CONTAINER(ds_dialog),box);
  
  bbox = gtk_hbox_new(1,dialog_button_spacing);
  gtk_widget_show(bbox);
  gtk_box_pack_end(GTK_BOX(box),bbox,0,0,0);

  if ( game_get_option_value(the_game,GOFlowers,NULL) )
    w = gtk_label_new("Declare Flowers/Seasons (and kongs)\nSelect tile and:");
  else
    w = gtk_label_new("Declare kongs\nSelect tile and:");
  gtk_widget_show(w);
  gtk_box_pack_end(GTK_BOX(box),w,0,0,0);
  
  ds_accel = gtk_accel_group_new();

  if ( game_get_option_value(the_game,GOFlowers,NULL) ) {
    ds_dialog_declare_button = w = gtk_button_new_with_label("Declare");
    gtk_widget_show(w);
    gtk_box_pack_start(GTK_BOX(bbox),w,1,1,0);
    gtk_signal_connect(GTK_OBJECT(w),"clicked",
		       turn_callback,(gpointer)PMsgDeclareSpecial);
    gtk_accel_group_add(ds_accel,GDK_d,0,0,GTK_OBJECT(w),"clicked");
    gtk_label_set_pattern(GTK_LABEL(GTK_BIN(w)->child),"_");
  }

  w = gtk_button_new_with_label("Kong");
  gtk_widget_show(w);
  gtk_box_pack_start(GTK_BOX(bbox),w,1,1,0);
  gtk_signal_connect(GTK_OBJECT(w),"clicked",
		     turn_callback,(gpointer)PMsgDeclareClosedKong);
  gtk_accel_group_add(ds_accel,GDK_k,0,0,GTK_OBJECT(w),"clicked");
  gtk_label_set_pattern(GTK_LABEL(GTK_BIN(w)->child),"_");

  ds_dialog_finish_button = w = gtk_button_new_with_label("Finish");
  gtk_widget_show(w);
  gtk_box_pack_start(GTK_BOX(bbox),w,1,1,0);
  gtk_signal_connect(GTK_OBJECT(w),"clicked",
		     turn_callback,(gpointer)PMsgNoClaim);
  gtk_accel_group_add(ds_accel,GDK_f,0,0,GTK_OBJECT(w),"clicked");
  gtk_label_set_pattern(GTK_LABEL(GTK_BIN(w)->child),"_");

  gtk_signal_connect(GTK_OBJECT(ds_dialog),"hide",
    add_or_remove_ds_accels,(gpointer)0);
  gtk_signal_connect(GTK_OBJECT(ds_dialog),"show",
    add_or_remove_ds_accels,(gpointer)1);
}

void ds_dialog_popup(void) {
  dialog_popup(ds_dialog,DPCentredOnce);
  if ( game_get_option_value(the_game,GOFlowers,NULL) )
    gtk_widget_grab_focus(ds_dialog_declare_button);
  else 
    gtk_widget_grab_focus(ds_dialog_finish_button);
}
/* dialog to ask to continue with next hand */
static GtkWidget *continue_dialog_continue_button;
static GtkWidget *continue_dialog_label;

void continue_dialog_init(void) {
  GtkWidget *box, *w;

  if ( continue_dialog ) {
    gtk_widget_destroy(continue_dialog);
    continue_dialog = NULL;
  }

  switch ( dialogs_position ) {
  case DialogsCentral:
  case DialogsUnspecified:
    continue_dialog = gtk_event_box_new();
    gtk_fixed_put(GTK_FIXED(discard_area),continue_dialog,0,0);
    break;
  case DialogsBelow:
    continue_dialog = gtk_event_box_new();
    gtk_box_pack_start(GTK_BOX(dialoglowerbox),continue_dialog,1,0,0);
    break;
  case DialogsPopup:
    continue_dialog = gtk_window_new(GTK_WINDOW_DIALOG);
    gtk_signal_connect (GTK_OBJECT (continue_dialog), "delete_event",
                               GTK_SIGNAL_FUNC (gtk_widget_hide), NULL);
  }
  
  box = gtk_vbox_new(0,dialog_vert_spacing);
  gtk_container_set_border_width(GTK_CONTAINER(box),
				 dialog_border_width);
  gtk_widget_show(box);
  gtk_container_add(GTK_CONTAINER(continue_dialog),box);
  
  continue_dialog_continue_button = w = gtk_button_new_with_label("Ready");
  gtk_widget_show(w);
  gtk_box_pack_end(GTK_BOX(box),w,0,0,0);
  gtk_signal_connect(GTK_OBJECT(w),"clicked",
		     continue_callback,(gpointer)PMsgDiscard);

  /* created and packed second so things float to bottom */

  continue_dialog_label = w = gtk_label_new("Here is some dummy text to space");
  gtk_widget_show(w);
  gtk_box_pack_end(GTK_BOX(box),w,0,0,0);
  
}

void continue_dialog_popup(void) {
  static char buf[256];
  /* the text of the display depends on whether we've said we're ready */
  if ( ! the_game->active ) {
    strcpy(buf,"Waiting for game to (re)start");
    gtk_label_set_text(GTK_LABEL(continue_dialog_label),buf);
    gtk_widget_hide(continue_dialog_continue_button);
  }
  else if ( the_game->ready[our_seat] ) {
    strcpy(buf,"Waiting for others ");
    strcat(buf,the_game->paused);
    gtk_label_set_text(GTK_LABEL(continue_dialog_label),buf);
    gtk_widget_hide(continue_dialog_continue_button);
  } else {
    strcpy(buf,"Ready ");
    strcat(buf,the_game->paused);
    strcat(buf,"?");
    gtk_label_set_text(GTK_LABEL(continue_dialog_label),buf);
    gtk_widget_show(continue_dialog_continue_button);
    gtk_widget_grab_focus(continue_dialog_continue_button);
  }
  dialog_popup(continue_dialog,DPCentredOnce);
}

static void continue_callback(GtkWidget *w UNUSED, gpointer data UNUSED) {
  PMsgReadyMsg rm;
  rm.type = PMsgReady;
  send_packet(&rm);
}

/* how many are up; used for positioning - not multi-thread safe! */
static int num_error_dialogs = 0; 

/* function called to close an error dialog */
static void kill_error(GtkObject *data) {
  gtk_widget_destroy(GTK_WIDGET(data));
  num_error_dialogs--;
}

/* popup an error box (new for each message),
   or an info box (which doesn't have error tile in it)' */
static void _error_dialog_popup(char *msg,int iserror) {
  GtkWidget *box, *hbox, *w, *error_dialog, *error_message;
  
  if ( num_error_dialogs >= 10 ) {
    warn("Too many error dialogs up to print error: %s",msg);
    return;
  }

  error_dialog = gtk_window_new(GTK_WINDOW_DIALOG);
  gtk_signal_connect_object(GTK_OBJECT (error_dialog), "delete_event",
                               GTK_SIGNAL_FUNC (gtk_widget_destroy), 
			    GTK_OBJECT(error_dialog));
  gtk_container_set_border_width(GTK_CONTAINER(error_dialog),
				 dialog_border_width);
  
  box = gtk_vbox_new(0,dialog_vert_spacing);
  gtk_widget_show(box);
  gtk_container_add(GTK_CONTAINER(error_dialog),box);
  
  hbox = gtk_hbox_new(0,dialog_button_spacing);
  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(box),hbox,0,0,0);

  if ( iserror ) {
    w = gtk_pixmap_new(tilepixmaps[0][ErrorTile],NULL);
    gtk_widget_show(w);
    gtk_box_pack_start(GTK_BOX(hbox),w,0,0,0);
  }

  error_message = gtk_label_new(msg);
  gtk_widget_show(error_message);
  gtk_box_pack_start(GTK_BOX(hbox),error_message,0,0,0);

  w = gtk_button_new_with_label("OK");
  gtk_widget_show(w);
  gtk_box_pack_start(GTK_BOX(box),w,0,0,0);
  gtk_signal_connect_object(GTK_OBJECT(w),"clicked",
		     kill_error,GTK_OBJECT(error_dialog));
  gtk_window_set_focus(GTK_WINDOW(error_dialog),w);

  dialog_popup(error_dialog,DPErrorPos);
  num_error_dialogs++;
  gdk_window_raise(error_dialog->window);
}

void error_dialog_popup(char *msg) {
  _error_dialog_popup(msg,1);
}

void info_dialog_popup(char *msg) {
  _error_dialog_popup(msg,0);
}



GtkWidget *openfile, *openhost, *openport,
  *openfiletext, *openhosttext, *openporttext, *openidtext, *opennametext,
  *opengamefile, *opengamefiletext, *opentimeout;

GtkWidget *openallowdisconnectbutton,*opensaveonexitbutton,*openrandomseatsbutton,
  *openplayercheckboxes[3],*openplayernames[3],*openplayeroptions[3],*opentimeoutspinbutton;
static GtkWidget *opengamepanel,*openplayeroptionboxes[3],
  *openconnectbutton,*openstartbutton, *openresumebutton;

/* callback used in next function */
static void openbut_callback(GtkWidget *w, gpointer data) {
  int active = GTK_TOGGLE_BUTTON(w)->active;

  switch ( (int)data ) {
  case 1:
    if ( active ) {
      /* selected network */
      gtk_widget_set_sensitive(openfile,0);
      gtk_widget_set_sensitive(openhost,1);
      gtk_widget_set_sensitive(openport,1);
    }
    break;
  case 2:
    if ( active ) {
      /* selected Unix socket */
      gtk_widget_set_sensitive(openfile,1);
      gtk_widget_set_sensitive(openhost,0);
      gtk_widget_set_sensitive(openport,0);
    }
    break;
  case 10:
  case 11:
  case 12:
    gtk_widget_set_sensitive(openplayeroptionboxes[((int)data)-10],active);
    break;
  }
}

/* the dialog for opening a connection.
   Arguments are initial values for host, port, file, id, name 
   text entry fields, true if use host or false for socket */
void open_dialog_init(char *ht, char *pt, char *ft, char *idt, char *nt, int usehost) {
  /* Layout of the box:
     x  Connect to host
     x  Use Unix socket
     Hostname
     ........
     Port 
     ....
     Filename
     ....
     Player ID
     ....
     Name
     ....
     OPEN CANCEL
  */
  int i;
  GtkWidget *box, *bbox, *but1, *but2, *w1, *w2;
  open_dialog = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  /* This one is allowed to shrink, and should */
  gtk_window_set_policy(GTK_WINDOW(open_dialog),1,1,1);
  gtk_signal_connect (GTK_OBJECT (open_dialog), "delete_event",
                               GTK_SIGNAL_FUNC (gtk_widget_hide), NULL);
  gtk_container_set_border_width(GTK_CONTAINER(open_dialog),
				 dialog_border_width);
  
  box = gtk_vbox_new(0,dialog_vert_spacing);
  gtk_widget_show(box);
  gtk_container_add(GTK_CONTAINER(open_dialog),box);
  
  w1 = gtk_hbox_new(0,dialog_button_spacing);
  gtk_widget_show(w1);
  gtk_box_pack_start(GTK_BOX(box),w1,0,0,0);
  but1 = gtk_radio_button_new_with_label(NULL,"Internet server");
  GTK_WIDGET_UNSET_FLAGS(but1,GTK_CAN_FOCUS);
#ifndef WIN32
  gtk_widget_show(but1);
#endif
  gtk_box_pack_start(GTK_BOX(w1),but1,0,0,0);
  but2 = gtk_radio_button_new_with_label(gtk_radio_button_group(GTK_RADIO_BUTTON(but1)),"Unix socket server");
  GTK_WIDGET_UNSET_FLAGS(but2,GTK_CAN_FOCUS);
#ifndef WIN32
  gtk_widget_show(but2);
#endif
  gtk_box_pack_end(GTK_BOX(w1),but2,0,0,0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(usehost? but1 : but2),1);

  gtk_signal_connect(GTK_OBJECT(but1),"toggled",openbut_callback,(gpointer)1);
  gtk_signal_connect(GTK_OBJECT(but2),"toggled",openbut_callback,(gpointer)2);

  w2 = gtk_hbox_new(0,dialog_button_spacing);
  gtk_widget_show(w2);
  gtk_box_pack_start(GTK_BOX(box),w2,0,0,0);
  openhost = gtk_hbox_new(0,0);
  gtk_widget_show(openhost);
  gtk_widget_set_sensitive(openhost,usehost);
  gtk_box_pack_start(GTK_BOX(w2),openhost,0,0,0);
  w1 = gtk_label_new("Host: ");
  gtk_widget_show(w1);
  gtk_box_pack_start(GTK_BOX(openhost),w1,0,0,0);
  openhosttext = gtk_entry_new();
  gtk_widget_show(openhosttext);
  gtk_entry_set_text(GTK_ENTRY(openhosttext),ht);
  gtk_box_pack_start(GTK_BOX(openhost),openhosttext,0,0,0);

  openport = gtk_hbox_new(0,0);
  gtk_widget_show(openport);
  gtk_widget_set_sensitive(openport,usehost);
  gtk_box_pack_end(GTK_BOX(w2),openport,0,0,0);
  w1 = gtk_label_new("Port: ");
  gtk_widget_show(w1);
  gtk_box_pack_start(GTK_BOX(openport),w1,0,0,0);
  openporttext = gtk_entry_new_with_max_length(5);
  gtk_widget_set_usize(openporttext,75,0);
  gtk_widget_show(openporttext);
  gtk_entry_set_text(GTK_ENTRY(openporttext),pt);
  gtk_box_pack_start(GTK_BOX(openport),openporttext,0,0,0);

  openfile = gtk_hbox_new(0,0);
#ifndef WIN32
  gtk_widget_show(openfile);
#endif
  gtk_widget_set_sensitive(openfile,!usehost);
  gtk_box_pack_start(GTK_BOX(box),openfile,0,0,0);
  w1 = gtk_label_new("Socket file: ");
  gtk_widget_show(w1);
  gtk_box_pack_start(GTK_BOX(openfile),w1,0,0,0);
  openfiletext = gtk_entry_new();
  gtk_widget_show(openfiletext);
  gtk_entry_set_text(GTK_ENTRY(openfiletext),ft);
  gtk_box_pack_start(GTK_BOX(openfile),openfiletext,0,0,0);

  opengamefile = gtk_hbox_new(0,0);
  gtk_widget_show(opengamefile);
  gtk_box_pack_start(GTK_BOX(box),opengamefile,0,0,0);
  w1 = gtk_label_new("Saved game file: ");
  gtk_widget_show(w1);
  gtk_box_pack_start(GTK_BOX(opengamefile),w1,0,0,0);
  opengamefiletext = gtk_entry_new();
  gtk_widget_show(opengamefiletext);
  gtk_entry_set_text(GTK_ENTRY(opengamefiletext),ft);
  gtk_box_pack_start(GTK_BOX(opengamefile),opengamefiletext,0,0,0);

  w2 = gtk_hbox_new(0,0);
  gtk_widget_show(w2);
  gtk_box_pack_start(GTK_BOX(box),w2,0,0,0);
  w1 = gtk_label_new("Player ID: ");
  gtk_widget_show(w1);
  gtk_box_pack_start(GTK_BOX(w2),w1,0,0,0);
  openidtext = gtk_entry_new();
  gtk_widget_show(openidtext);
  gtk_entry_set_text(GTK_ENTRY(openidtext),idt);
  gtk_box_pack_start(GTK_BOX(w2),openidtext,0,0,0);

  w2 = gtk_hbox_new(0,0);
  gtk_widget_show(w2);
  gtk_box_pack_start(GTK_BOX(box),w2,0,0,0);
  w1 = gtk_label_new("Name: ");
  gtk_widget_show(w1);
  gtk_box_pack_start(GTK_BOX(w2),w1,0,0,0);
  opennametext = gtk_entry_new();
  gtk_widget_show(opennametext);
  if ( ! nt ) nt = getenv("LOGNAME");
  if ( ! nt ) nt = getlogin(); /* may need to be in sysdep.c */
  gtk_entry_set_text(GTK_ENTRY(opennametext),nt);
  gtk_box_pack_start(GTK_BOX(w2),opennametext,0,0,0);

  /* Now some stuff for when this panel is in its personality as
     a start game panel */
  opengamepanel = bbox = gtk_vbox_new(0,dialog_vert_spacing);
  /* gtk_widget_show(bbox); */
  gtk_box_pack_start(GTK_BOX(box),bbox,0,0,0);

  for ( i = 0 ; i < 3 ; i++ ) {
    static const char *playerlabs[] = { "Second player:" , "Third player:",
					"Fourth player:" };

    w2 = gtk_hbox_new(0,dialog_button_spacing);
    gtk_widget_show(w2);
    gtk_box_pack_start(GTK_BOX(bbox),w2,0,0,0);
    w1 = gtk_label_new(playerlabs[i]);
    gtk_widget_show(w1);
    gtk_box_pack_start(GTK_BOX(w2),w1,0,0,0);
    openplayercheckboxes[i] = w1 = 
      gtk_check_button_new_with_label("Start computer player   ");
    gtk_widget_show(w1);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w1),1);
    gtk_box_pack_end(GTK_BOX(w2),w1,0,0,0);
    gtk_signal_connect(GTK_OBJECT(w1),"toggled",openbut_callback,(gpointer)(10+i));
    openplayeroptionboxes[i] = w2 = gtk_hbox_new(0,dialog_button_spacing);
    gtk_widget_show(w2);
    gtk_box_pack_start(GTK_BOX(bbox),w2,0,0,0);
    w1 = gtk_label_new(" Name:");
    gtk_widget_show(w1);
    gtk_box_pack_start(GTK_BOX(w2),w1,0,0,0);
    openplayernames[i] = w1 = gtk_entry_new();
    gtk_entry_set_text(GTK_ENTRY(w1),robot_names[i]);
    gtk_widget_show(w1);
    gtk_box_pack_start(GTK_BOX(w2),w1,0,0,0);
    w1 = gtk_label_new("  Options:");
    gtk_widget_show(w1);
    gtk_box_pack_start(GTK_BOX(w2),w1,0,0,0);
    openplayeroptions[i] = w1 = gtk_entry_new();
    gtk_widget_show(w1);
    gtk_box_pack_start(GTK_BOX(w2),w1,0,0,0);
  }

  openallowdisconnectbutton = w1 =
    gtk_check_button_new_with_label("Allow disconnection from game");
  gtk_widget_show(w1);
  gtk_box_pack_start(GTK_BOX(bbox),w1,0,0,0);

  opensaveonexitbutton = w1 =
    gtk_check_button_new_with_label("Save game state on exit");
  gtk_widget_show(w1);
  gtk_box_pack_start(GTK_BOX(bbox),w1,0,0,0);

  openrandomseatsbutton = w1 =
    gtk_check_button_new_with_label("Seat players randomly");
  gtk_widget_show(w1);
  gtk_box_pack_start(GTK_BOX(bbox),w1,0,0,0);

  opentimeoutspinbutton = w1 = 
    gtk_spin_button_new(GTK_ADJUSTMENT(gtk_adjustment_new(1.0*game_get_option_entry_from_table(&prefs_table,GOTimeout,NULL)->value.optnat,0.0,300.0,1.0,10.0,0.0)),
			0.0,0);
  gtk_widget_show(w1);
  opentimeout = w2 = gtk_hbox_new(0,dialog_button_spacing);
  gtk_widget_show(w2);
  gtk_box_pack_start(GTK_BOX(w2),w1,0,0,0);
  w1 = gtk_label_new("seconds allowed for claims");
  gtk_widget_show(w1);
  gtk_box_pack_start(GTK_BOX(w2),w1,0,0,0);
  gtk_box_pack_start(GTK_BOX(bbox),w2,0,0,0);

  w1 = gtk_hbox_new(0,dialog_button_spacing);
  gtk_widget_show(w1);
  gtk_box_pack_start(GTK_BOX(box),w1,0,0,0);
  openconnectbutton = w2 = gtk_button_new_with_label("Connect");
  gtk_widget_show(w2);
  gtk_box_pack_start(GTK_BOX(w1),w2,0,0,0);
  gtk_signal_connect(GTK_OBJECT(w2),"clicked",open_connection,0);
  openstartbutton = w2 = gtk_button_new_with_label("Start Game");
  /* gtk_widget_show(w2); */
  gtk_box_pack_start(GTK_BOX(w1),w2,0,0,0);
  gtk_signal_connect(GTK_OBJECT(w2),"clicked",open_connection,(gpointer)1);
  openresumebutton = w2 = gtk_button_new_with_label("Resume Game");
  /* gtk_widget_show(w2); */
  gtk_box_pack_start(GTK_BOX(w1),w2,0,0,0);
  gtk_signal_connect(GTK_OBJECT(w2),"clicked",open_connection,(gpointer)2);
  w2 = gtk_button_new_with_label("Cancel");
  gtk_widget_show(w2);
  gtk_box_pack_end(GTK_BOX(w1),w2,0,0,0);
  gtk_signal_connect_object(GTK_OBJECT(w2),"clicked",
			    close_saving_posn,GTK_OBJECT(open_dialog));
}

void open_dialog_popup(GtkWidget *w UNUSED, gpointer data) {
  int new = 0, join = 0, resume = 0;
  /* set the default id to be our current id */
  char buf[256];
  sprintf(buf,"%d",our_id);

  if ( (int)data == 1 ) new = 1;
  if ( (int)data == 0 ) join = 1;
  if ( (int)data == 2 ) resume = 1;
  gtk_entry_set_text(GTK_ENTRY(openidtext),buf);
  if ( resume ) gtk_widget_show(opengamefile);
  else gtk_widget_hide(opengamefile);
  if ( new || resume ) gtk_widget_show(opengamepanel);
  else gtk_widget_hide(opengamepanel);
  if ( new ) gtk_widget_show(opentimeout);
  else gtk_widget_hide(opentimeout);
  if ( new ) gtk_widget_show(openrandomseatsbutton);
  else gtk_widget_hide(openrandomseatsbutton);
  if ( new ) gtk_widget_show(openstartbutton);
  else gtk_widget_hide(openstartbutton);
  if ( join ) gtk_widget_show(openconnectbutton);
  else gtk_widget_hide(openconnectbutton);
  if ( resume ) gtk_widget_show(openresumebutton);
  else gtk_widget_hide(openresumebutton);

  dialog_popup(open_dialog,DPCentredOnce);

  if ( new ) gtk_widget_grab_focus(openstartbutton);
  if ( join ) gtk_widget_grab_focus(openconnectbutton);
  if ( resume ) gtk_widget_grab_focus(openresumebutton);
}


/* an accelerator group  for the turn dialog */
static GtkAccelGroup *turn_accel;

/* this little function is used to install or remove the
   turn accelerators */
static void add_or_remove_turn_accels(GtkWidget *w UNUSED, gpointer data)
{
  if ( data )
    gtk_window_add_accel_group(GTK_WINDOW(topwindow),turn_accel);
  else
    gtk_window_remove_accel_group(GTK_WINDOW(topwindow),turn_accel);
}

/* the turn dialog: buttons for Discard (also declares specs),
   Kong (concealed, of course), Add to Pung, Mah Jong */
void turn_dialog_init(void) {
  GtkWidget *box, *butbox, *w;

  if ( turn_dialog ) {
    gtk_widget_destroy(turn_dialog);
    turn_dialog = NULL;
  }

  switch ( dialogs_position ) {
  case DialogsCentral:
  case DialogsUnspecified:
    turn_dialog = gtk_event_box_new();
    gtk_fixed_put(GTK_FIXED(discard_area),turn_dialog,0,0);
    break;
  case DialogsBelow:
    turn_dialog = gtk_event_box_new();
    gtk_box_pack_start(GTK_BOX(dialoglowerbox),turn_dialog,1,0,0);
    break;
  case DialogsPopup:
    turn_dialog = gtk_window_new(GTK_WINDOW_DIALOG);
    gtk_signal_connect (GTK_OBJECT (turn_dialog), "delete_event",
                               GTK_SIGNAL_FUNC (gtk_widget_hide), NULL);
  }

  box = gtk_vbox_new(0,dialog_vert_spacing);
  gtk_container_set_border_width(GTK_CONTAINER(box),
				 dialog_border_width);
  gtk_widget_show(box);
  gtk_container_add(GTK_CONTAINER(turn_dialog),box);

  butbox = gtk_hbox_new(1,dialog_button_spacing);
  gtk_widget_show(butbox);
  gtk_box_pack_end(GTK_BOX(box),butbox,0,0,0);
    
  turn_dialog_label = w = gtk_label_new("Select tile and:");
  gtk_widget_show(w);
  gtk_box_pack_end(GTK_BOX(box),w,0,0,0);

  turn_accel = gtk_accel_group_new();

  w = gtk_button_new_with_label("Discard");
  gtk_widget_show(w);
  gtk_box_pack_start(GTK_BOX(butbox),w,1,1,0);
  /* so other function can set it */
  turn_dialog_discard_button = w;
  gtk_signal_connect(GTK_OBJECT(w),"clicked",
		     turn_callback,(gpointer)PMsgDiscard);
  gtk_accel_group_add(turn_accel,GDK_d,0,0,GTK_OBJECT(w),"clicked");
  gtk_label_set_pattern(GTK_LABEL(GTK_BIN(w)->child),"_");
    
  w = gtk_button_new_with_label("& Calling");
  gtk_widget_show(w);
  gtk_box_pack_start(GTK_BOX(butbox),w,1,1,0);
  /* so other function can set it */
  turn_dialog_calling_button = w;
  /* this assumes knowledge that protocol enums don't go above
     1000000 */
  gtk_signal_connect(GTK_OBJECT(w),"clicked",
		     turn_callback,(gpointer)(PMsgDiscard+1000000));
  gtk_accel_group_add(turn_accel,GDK_c,0,0,GTK_OBJECT(w),"clicked");
  gtk_label_set_pattern(GTK_LABEL(GTK_BIN(w)->child),"  _");
    
  w = gtk_button_new_with_label("Kong");
  GTK_WIDGET_UNSET_FLAGS(w,GTK_CAN_FOCUS);
  gtk_widget_show(w);
  gtk_box_pack_start(GTK_BOX(butbox),w,1,1,0);
  gtk_signal_connect(GTK_OBJECT(w),"clicked",
		     turn_callback,(gpointer)PMsgDeclareClosedKong);
  gtk_accel_group_add(turn_accel,GDK_k,0,0,GTK_OBJECT(w),"clicked");
  gtk_label_set_pattern(GTK_LABEL(GTK_BIN(w)->child),"_");
    
  w = gtk_button_new_with_label("Add to Pung");
  GTK_WIDGET_UNSET_FLAGS(w,GTK_CAN_FOCUS);
  gtk_widget_show(w);
  gtk_box_pack_start(GTK_BOX(butbox),w,1,1,0);
  gtk_signal_connect(GTK_OBJECT(w),"clicked",
		     turn_callback,(gpointer)PMsgAddToPung);
  gtk_accel_group_add(turn_accel,GDK_p,0,0,GTK_OBJECT(w),"clicked");
  gtk_label_set_pattern(GTK_LABEL(GTK_BIN(w)->child),"       _");
    
  w = gtk_button_new_with_label("Mah Jong!");
  GTK_WIDGET_UNSET_FLAGS(w,GTK_CAN_FOCUS);
  gtk_widget_show(w);
  gtk_box_pack_start(GTK_BOX(butbox),w,1,1,0);
  gtk_signal_connect(GTK_OBJECT(w),"clicked",
		     turn_callback,(gpointer)PMsgMahJong);
  gtk_accel_group_add(turn_accel,GDK_m,0,0,GTK_OBJECT(w),"clicked");
  gtk_label_set_pattern(GTK_LABEL(GTK_BIN(w)->child),"_");

  gtk_signal_connect(GTK_OBJECT(turn_dialog),"hide",
    add_or_remove_turn_accels,(gpointer)0);
  gtk_signal_connect(GTK_OBJECT(turn_dialog),"show",
    add_or_remove_turn_accels,(gpointer)1);

}
      
/* dialog for scoring phase: forming closed sets */
static GtkWidget *scoring_done, *scoring_special;

static GtkAccelGroup *scoring_accel;

static void add_or_remove_scoring_accels(GtkWidget *w UNUSED, gpointer data)
{
  if ( data )
    gtk_window_add_accel_group(GTK_WINDOW(topwindow),scoring_accel);
  else
    gtk_window_remove_accel_group(GTK_WINDOW(topwindow),scoring_accel);
}


void scoring_dialog_init(void) {
  GtkWidget *box, *butbox, *w;
  
  if ( scoring_dialog ) {
    gtk_widget_destroy(scoring_dialog);
    scoring_dialog = NULL;
  }

  switch ( dialogs_position ) {
  case DialogsCentral:
  case DialogsUnspecified:
    scoring_dialog = gtk_event_box_new();
    gtk_fixed_put(GTK_FIXED(discard_area),scoring_dialog,0,0);
    break;
  case DialogsBelow:
    scoring_dialog = gtk_event_box_new();
    gtk_box_pack_start(GTK_BOX(dialoglowerbox),scoring_dialog,1,0,0);
    break;
  case DialogsPopup:
    scoring_dialog = gtk_window_new(GTK_WINDOW_DIALOG);
    gtk_signal_connect (GTK_OBJECT (scoring_dialog), "delete_event",
                               GTK_SIGNAL_FUNC (gtk_widget_hide), NULL);
  }
  
  box = gtk_vbox_new(0,dialog_vert_spacing);
  gtk_container_set_border_width(GTK_CONTAINER(box),
				 dialog_border_width);
  gtk_widget_show(box);
  gtk_container_add(GTK_CONTAINER(scoring_dialog),box);

  butbox = gtk_hbox_new(1,dialog_button_spacing);
  gtk_widget_show(butbox);
  gtk_box_pack_end(GTK_BOX(box),butbox,0,0,0);
  
  w = gtk_label_new("Declare concealed sets\nSelect 1st tile and:");
  gtk_widget_show(w);
  gtk_box_pack_end(GTK_BOX(box),w,0,0,0);

  scoring_accel = gtk_accel_group_new();

  w = gtk_button_new_with_label("Eyes");
  GTK_WIDGET_UNSET_FLAGS(w,GTK_CAN_FOCUS);
  gtk_widget_show(w);
  gtk_box_pack_start(GTK_BOX(butbox),w,1,1,0);
  gtk_signal_connect(GTK_OBJECT(w),"clicked",
		     turn_callback,(gpointer)PMsgFormClosedPair);
  gtk_accel_group_add(scoring_accel,GDK_e,0,0,GTK_OBJECT(w),"clicked");
  gtk_label_set_pattern(GTK_LABEL(GTK_BIN(w)->child),"_");

  
  w = gtk_button_new_with_label("Chow");
  GTK_WIDGET_UNSET_FLAGS(w,GTK_CAN_FOCUS);
  gtk_widget_show(w);
  gtk_box_pack_start(GTK_BOX(butbox),w,1,1,0);
  gtk_signal_connect(GTK_OBJECT(w),"clicked",
		     turn_callback,(gpointer)PMsgFormClosedChow);
  gtk_accel_group_add(scoring_accel,GDK_c,0,0,GTK_OBJECT(w),"clicked");
  gtk_label_set_pattern(GTK_LABEL(GTK_BIN(w)->child),"_");
  
  w = gtk_button_new_with_label("Pung");
  GTK_WIDGET_UNSET_FLAGS(w,GTK_CAN_FOCUS);
  gtk_widget_show(w);
  gtk_box_pack_start(GTK_BOX(butbox),w,1,1,0);
  gtk_signal_connect(GTK_OBJECT(w),"clicked",
		     turn_callback,(gpointer)PMsgFormClosedPung);
  gtk_accel_group_add(scoring_accel,GDK_p,0,0,GTK_OBJECT(w),"clicked");
  gtk_label_set_pattern(GTK_LABEL(GTK_BIN(w)->child),"_");
  
  scoring_special = w = gtk_button_new_with_label("Special Hand");
  GTK_WIDGET_UNSET_FLAGS(w,GTK_CAN_FOCUS);
  gtk_widget_show(w);
  gtk_box_pack_start(GTK_BOX(butbox),w,1,1,0);
  gtk_signal_connect(GTK_OBJECT(w),"clicked",
		     turn_callback,(gpointer)PMsgFormClosedSpecialSet);
  gtk_accel_group_add(scoring_accel,GDK_s,0,0,GTK_OBJECT(w),"clicked");
  gtk_label_set_pattern(GTK_LABEL(GTK_BIN(w)->child),"_");
  
  scoring_done = w = gtk_button_new_with_label("Finished");
  GTK_WIDGET_UNSET_FLAGS(w,GTK_CAN_FOCUS);
  /* gtk_widget_show(w); */ /* uses same space as special hand */
  gtk_box_pack_start(GTK_BOX(butbox),w,1,1,0);
  gtk_signal_connect(GTK_OBJECT(w),"clicked",
		     turn_callback,(gpointer)PMsgShowTiles);
  gtk_accel_group_add(scoring_accel,GDK_f,0,0,GTK_OBJECT(w),"clicked");
  gtk_label_set_pattern(GTK_LABEL(GTK_BIN(w)->child),"_");

  gtk_signal_connect(GTK_OBJECT(scoring_dialog),"hide",
    add_or_remove_scoring_accels,(gpointer)0);
  gtk_signal_connect(GTK_OBJECT(scoring_dialog),"show",
    add_or_remove_scoring_accels,(gpointer)1);

}

void scoring_dialog_popup(void) {
  if ( the_game->player == our_seat ) {
    gtk_widget_show(scoring_special);
    gtk_widget_hide(scoring_done);
  } else {
    gtk_widget_hide(scoring_special);
    gtk_widget_show(scoring_done);
  }
  dialog_popup(scoring_dialog,DPCentredOnce);
}



/* close a widget, saving its position for next open */
void close_saving_posn(GtkWidget *w) {
  gint x,y;
  gdk_window_get_deskrelative_origin(w->window,&x,&y);
  gtk_widget_set_uposition(w,x,y);
  gtk_widget_hide(w);
}

/* the textwindow for scoring information etc */
void textwindow_init(void) {
  int i;
  GtkWidget *sbar, *obox, *box, *tmp, *lbl,*textw;

  textwindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_signal_connect (GTK_OBJECT (textwindow), "delete_event",
                               GTK_SIGNAL_FUNC (gtk_widget_hide), NULL);
  /* must allow shrinking */
  gtk_window_set_policy(GTK_WINDOW(textwindow),TRUE,TRUE,FALSE);
  gtk_window_set_title(GTK_WINDOW(textwindow),"Scoring calculations");
  /* reasonable size is depends on what fonts come up by... this seems to work */
  /* we now set homogeneous tabs, which forces it pretty wide anyway */
#ifdef WIN32
  gtk_window_set_default_size(GTK_WINDOW(textwindow),000,400);
#else
  gtk_window_set_default_size(GTK_WINDOW(textwindow),000,300);
#endif
  obox = gtk_vbox_new(0,dialog_vert_spacing);
  gtk_container_set_border_width(GTK_CONTAINER(obox),dialog_border_width);
  gtk_widget_show(obox);
  gtk_container_add(GTK_CONTAINER(textwindow),obox);

  scoring_notebook = gtk_notebook_new();
  GTK_WIDGET_UNSET_FLAGS(scoring_notebook,GTK_CAN_FOCUS);
  gtk_notebook_set_homogeneous_tabs(GTK_NOTEBOOK(scoring_notebook),1);
  gtk_widget_show(scoring_notebook);
  gtk_box_pack_start(GTK_BOX(obox),scoring_notebook,1,1,0);
  
  for (i=0;i<5;i++) {
    GtkStyle *s;
    textw = gtk_text_new(NULL,NULL);
    gtk_widget_show(textw);
    textpages[i] = textw;
    sbar = gtk_vscrollbar_new(GTK_TEXT (textw)->vadj);
    gtk_widget_show(sbar);
    box = gtk_hbox_new(0,0);
    gtk_widget_show(box);
    gtk_box_pack_start(GTK_BOX(box),sbar,0,0,0);
    gtk_box_pack_start(GTK_BOX(box),textw,1,1,0);
    lbl = gtk_label_new((i==4) ? "Settlement" : "");
    gtk_widget_show(lbl);
    textlabels[i] = lbl;
    gtk_notebook_append_page(GTK_NOTEBOOK(scoring_notebook),box,lbl);
    gtk_widget_realize(textw);
    if ( fixed_font ) {
      s = gtk_style_copy(gtk_widget_get_style(textw));
      s->font = fixed_font;
      gtk_widget_set_style(textw,s);
    }
  }

  tmp = gtk_button_new_with_label("Close");
  gtk_signal_connect_object(GTK_OBJECT(tmp),"clicked",
		     close_saving_posn,GTK_OBJECT(textwindow));
  gtk_widget_show(tmp);
  gtk_box_pack_end(GTK_BOX(obox),tmp,0,0,0);
  gtk_widget_grab_focus(tmp);
}

/* the callback when user hits return in the message composition window */
static void mentry_callback(GtkWidget *widget UNUSED,GtkWidget *mentry) {
  gchar *mentry_text;
  PMsgSendMessageMsg smm;
  mentry_text = gtk_entry_get_text(GTK_ENTRY(mentry));
  smm.type = PMsgSendMessage;
  smm.addressee = 0; /* broadcast only at present ... */
  smm.text = mentry_text;
  send_packet(&smm);
  gtk_entry_set_text(GTK_ENTRY(mentry),"");
}

/* the window for messages */
void messagewindow_init(void) {
  GtkWidget *sbar, *obox, *box, *tmp, *mentry, *label;
  GtkStyle *s;

  messagewindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_signal_connect (GTK_OBJECT (messagewindow), "delete_event",
                               GTK_SIGNAL_FUNC (gtk_widget_hide), NULL);
  /* must allow shrinking */
  gtk_window_set_policy(GTK_WINDOW(messagewindow),TRUE,TRUE,FALSE);
  gtk_window_set_title(GTK_WINDOW(messagewindow),"Messages");
  /* reasonable size is ... */
  gtk_window_set_default_size(GTK_WINDOW(messagewindow),400,300);

  obox = gtk_vbox_new(0,dialog_vert_spacing);
  gtk_container_set_border_width(GTK_CONTAINER(obox),dialog_border_width);
  gtk_widget_show(obox);
  gtk_container_add(GTK_CONTAINER(messagewindow),obox);

  messagetext = gtk_text_new(NULL,NULL);
  gtk_widget_show(messagetext);
  sbar = gtk_vscrollbar_new(GTK_TEXT (messagetext)->vadj);
  gtk_widget_show(sbar);
  box = gtk_hbox_new(0,0);
  gtk_widget_show(box);
  gtk_box_pack_start(GTK_BOX(box),sbar,0,0,0);
  gtk_box_pack_start(GTK_BOX(box),messagetext,1,1,0);
  gtk_box_pack_start(GTK_BOX(obox),box,1,1,0);
  gtk_widget_realize(messagetext);
  if ( fixed_font ) {
    s = gtk_style_copy(gtk_widget_get_style(messagetext));
    s->font = fixed_font;
    gtk_widget_set_style(messagetext,s);
  }

  GTK_WIDGET_UNSET_FLAGS(messagetext,GTK_CAN_FOCUS); /* entry widget shd focus */

  label = gtk_label_new("Send message:");
  gtk_misc_set_alignment(GTK_MISC(label),0.0,0.5);
  gtk_widget_show(label);
  gtk_box_pack_start(GTK_BOX(obox),label,0,0,0);
  GTK_WIDGET_UNSET_FLAGS(label,GTK_CAN_FOCUS);

  /* the thing for sending messages */
  mentry = gtk_entry_new();
  gtk_signal_connect(GTK_OBJECT(mentry),"activate",
		     GTK_SIGNAL_FUNC(mentry_callback),mentry);
  gtk_box_pack_start(GTK_BOX(obox),mentry,0,0,0);
  gtk_widget_show(mentry);

  tmp = gtk_button_new_with_label("Close");
  GTK_WIDGET_UNSET_FLAGS(tmp,GTK_CAN_FOCUS); /* entry widget shd focus */
  gtk_signal_connect_object(GTK_OBJECT(tmp),"clicked",
			    close_saving_posn,GTK_OBJECT(messagewindow));
  gtk_widget_show(tmp);
  gtk_box_pack_end(GTK_BOX(obox),tmp,0,0,0);
  gtk_widget_grab_focus(mentry);
}

/* callback for saving as */
static void save_callback(GtkWidget *w UNUSED, gpointer data UNUSED) {
  PMsgSaveStateMsg m;

  m.type = PMsgSaveState;
  m.filename = gtk_entry_get_text(GTK_ENTRY(save_text));
  send_packet(&m);
  close_saving_posn(save_window);
}

/* window for Save As ... function */
static void save_init(void) {
  GtkWidget *box, *bbox, *w1, *tmp;


  save_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_signal_connect (GTK_OBJECT (save_window), "delete_event",
                               GTK_SIGNAL_FUNC (gtk_widget_hide), NULL);
  gtk_container_set_border_width(GTK_CONTAINER(save_window),
				 dialog_border_width);

  bbox = gtk_vbox_new(0,0);
  gtk_widget_show(bbox);
  gtk_container_add(GTK_CONTAINER(save_window),bbox);

  w1 = gtk_label_new("Save as:");
  gtk_widget_show(w1);
  gtk_box_pack_start(GTK_BOX(bbox),w1,0,0,0);

  box = gtk_hbox_new(0,0);
  gtk_widget_show(box);
  gtk_box_pack_start(GTK_BOX(bbox),box,0,0,0);

  save_text = gtk_entry_new();
  gtk_widget_show(save_text);
  gtk_box_pack_start(GTK_BOX(box),save_text,0,0,0);

  w1 = gtk_label_new(".mjs");
  gtk_widget_show(w1);
  gtk_box_pack_start(GTK_BOX(box),w1,0,0,0);

  box = gtk_hbox_new(0,0);
  gtk_widget_show(box);
  gtk_box_pack_start(GTK_BOX(bbox),box,0,0,0);

  tmp = gtk_button_new_with_label("Save");
  gtk_signal_connect_object(GTK_OBJECT(tmp),"clicked",
		     save_callback,0);
  gtk_widget_show(tmp);
  gtk_box_pack_start(GTK_BOX(box),tmp,1,1,0);

  tmp = gtk_button_new_with_label("Cancel");
  gtk_signal_connect_object(GTK_OBJECT(tmp),"clicked",
		     close_saving_posn,GTK_OBJECT(save_window));
  gtk_widget_show(tmp);
  gtk_box_pack_start(GTK_BOX(box),tmp,1,1,0);

}

/* radio buttons for dialog positions */
static GtkWidget *display_option_dialog_dialogposn[DialogsPopup+1];
/* checkbox for animation */
static GtkWidget *display_option_dialog_animation;
/* checkbox for tiletips */
static GtkWidget *display_option_dialog_tiletips;
/* check box for iconify option */
static GtkWidget *display_option_dialog_iconify;
/* radiobuttons for show wall */
static GtkWidget *display_option_dialog_showwall[3];
static GtkWidget *display_option_dialog_tileset, *display_option_dialog_tileset_path;
/* names for players */
static GtkWidget *display_option_dialog_playernames[3];

static void display_option_dialog_apply(void);
static void display_option_dialog_save(void);
static void display_option_dialog_refresh(void);

/* window for display options */
static void display_option_dialog_init(void) {
  int i;
  GtkWidget *box, *bbox, *hbox, *but1, *but2, *but3, *w1, *w2;

  display_option_dialog = gtk_window_new(GTK_WINDOW_DIALOG);
  gtk_signal_connect (GTK_OBJECT (display_option_dialog), "delete_event",
                               GTK_SIGNAL_FUNC (gtk_widget_hide), NULL);
  gtk_container_set_border_width(GTK_CONTAINER(display_option_dialog),
				 dialog_border_width);

  box = gtk_vbox_new(0,dialog_vert_spacing);
  gtk_widget_show(box);
  gtk_container_add(GTK_CONTAINER(display_option_dialog),box);

  bbox = gtk_vbox_new(0,0);
  gtk_widget_show(bbox);
  gtk_box_pack_start(GTK_BOX(box),bbox,0,0,0);

  w1 = gtk_label_new("Position of action dialogs:");
  gtk_widget_show(w1);
  gtk_box_pack_start(GTK_BOX(bbox),w1,0,0,0);

  hbox = gtk_hbox_new(0,dialog_button_spacing);
  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(bbox),hbox,0,0,0);

  but1 = gtk_radio_button_new_with_label(NULL,"central");
  gtk_widget_show(but1);
  gtk_box_pack_start(GTK_BOX(hbox),but1,0,0,0);
  display_option_dialog_dialogposn[DialogsCentral] = but1 ;

  but2 = gtk_radio_button_new_with_label(gtk_radio_button_group(GTK_RADIO_BUTTON(but1)),"below");
  gtk_widget_show(but2);
  gtk_box_pack_start(GTK_BOX(hbox),but2,0,0,0);
  display_option_dialog_dialogposn[DialogsBelow] = but2;

  but3 = gtk_radio_button_new_with_label(gtk_radio_button_group(GTK_RADIO_BUTTON(but1)),"popup");
  gtk_widget_show(but3);
  gtk_box_pack_start(GTK_BOX(hbox),but3,0,0,0);
  display_option_dialog_dialogposn[DialogsPopup] = but3;

  /* animation */
  display_option_dialog_animation = but1 = 
    gtk_check_button_new_with_label("Animation");
  gtk_widget_show(but1);
  gtk_box_pack_start(GTK_BOX(box),but1,0,0,0);

  /* tiletips */
  display_option_dialog_tiletips = but1 = 
    gtk_check_button_new_with_label("Tiletips always shown");
  gtk_widget_show(but1);
  gtk_box_pack_start(GTK_BOX(box),but1,0,0,0);

  /* showwall setting */
  bbox = gtk_vbox_new(0,0);
  gtk_widget_show(bbox);
  gtk_box_pack_start(GTK_BOX(box),bbox,0,0,0);

  w1 = gtk_label_new("Show the wall:");
  gtk_widget_show(w1);
  gtk_box_pack_start(GTK_BOX(bbox),w1,0,0,0);

  hbox = gtk_hbox_new(0,dialog_button_spacing);
  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(bbox),hbox,0,0,0);

  but1 = gtk_radio_button_new_with_label(NULL,"always");
  gtk_widget_show(but1);
  gtk_box_pack_start(GTK_BOX(hbox),but1,0,0,0);
  display_option_dialog_showwall[1] = but1 ;

  but2 = gtk_radio_button_new_with_label(gtk_radio_button_group(GTK_RADIO_BUTTON(but1)),"when room");
  gtk_widget_show(but2);
  gtk_box_pack_start(GTK_BOX(hbox),but2,0,0,0);
  display_option_dialog_showwall[2] = but2;

  but3 = gtk_radio_button_new_with_label(gtk_radio_button_group(GTK_RADIO_BUTTON(but1)),"never");
  gtk_widget_show(but3);
  gtk_box_pack_start(GTK_BOX(hbox),but3,0,0,0);
  display_option_dialog_showwall[0] = but3;

#ifndef WIN32
  /* iconify */
  display_option_dialog_iconify = but1 = 
    gtk_check_button_new_with_label("Iconify all windows with main");
  gtk_widget_show(but1);
  gtk_box_pack_start(GTK_BOX(box),but1,0,0,0);
#endif

  /* tileset name */
  bbox = gtk_hbox_new(0,0);
  gtk_widget_show(bbox);
  gtk_box_pack_start(GTK_BOX(box),bbox,0,0,0);

  w1 = gtk_label_new("Tileset: ");
  gtk_widget_show(w1);
  gtk_box_pack_start(GTK_BOX(bbox),w1,0,0,0);
  
  display_option_dialog_tileset = w1 = gtk_entry_new();
  gtk_widget_show(w1);
  gtk_box_pack_start(GTK_BOX(bbox),w1,0,0,0);
  
  /* tileset path */
  bbox = gtk_hbox_new(0,0);
  gtk_widget_show(bbox);
  gtk_box_pack_start(GTK_BOX(box),bbox,0,0,0);

  w1 = gtk_label_new("Tileset Path: ");
  gtk_widget_show(w1);
  gtk_box_pack_start(GTK_BOX(bbox),w1,0,0,0);
  
  display_option_dialog_tileset_path = w1 = gtk_entry_new();
  gtk_widget_show(w1);
  gtk_box_pack_start(GTK_BOX(bbox),w1,0,0,0);

  /* robot player names */
  bbox = gtk_vbox_new(0,dialog_vert_spacing);
  gtk_widget_show(bbox);
  gtk_box_pack_start(GTK_BOX(box),bbox,0,0,0);

  w1 = gtk_label_new("Computer Player names:");
  gtk_widget_show(w1);
  gtk_box_pack_start(GTK_BOX(bbox),w1,0,0,0);
  
  for ( i = 0; i < 3; i++ ) {
    char buf[5];
    hbox = gtk_hbox_new(0,0);
    gtk_widget_show(hbox);
    gtk_box_pack_start(GTK_BOX(bbox),hbox,0,0,0);
    
    sprintf(buf,"%d: ",i+1);
    w1 = gtk_label_new(buf);
    gtk_widget_show(w1);
    gtk_box_pack_start(GTK_BOX(hbox),w1,0,0,0);
    
    display_option_dialog_playernames[i] = w1 = gtk_entry_new();
    gtk_widget_show(w1);
    gtk_box_pack_start(GTK_BOX(hbox),w1,0,0,0);
  }
  
  
  /* apply, save and close buttons */
  w1 = gtk_hbox_new(0,dialog_button_spacing);
  gtk_widget_show(w1);
  gtk_box_pack_end(GTK_BOX(box),w1,0,0,0);

  w2 = gtk_button_new_with_label("Done");
  gtk_widget_show(w2);
  gtk_box_pack_start(GTK_BOX(w1),w2,0,0,0);
  gtk_signal_connect(GTK_OBJECT(w2),"clicked",display_option_dialog_apply,NULL);

  w2 = gtk_button_new_with_label("Save");
  gtk_widget_show(w2);
  gtk_box_pack_start(GTK_BOX(w1),w2,0,0,0);
  gtk_signal_connect(GTK_OBJECT(w2),"clicked",display_option_dialog_save,NULL);

  w2 = gtk_button_new_with_label("Cancel");
  gtk_widget_show(w2);
  gtk_box_pack_start(GTK_BOX(w1),w2,0,0,0);
  gtk_signal_connect_object(GTK_OBJECT(w2),"clicked",gtk_widget_hide,GTK_OBJECT(display_option_dialog));
}

/* make the panel match reality */
static void display_option_dialog_refresh(void) {
  int i;
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(display_option_dialog_dialogposn[dialogs_position]),1);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(display_option_dialog_animation),animate);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(display_option_dialog_tiletips),tiletips);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(display_option_dialog_showwall[pref_showwall == -1 ? 2 : pref_showwall]),1);
#ifndef WIN32
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(display_option_dialog_iconify),iconify_dialogs_with_main);
#endif
  gtk_entry_set_text(GTK_ENTRY(display_option_dialog_tileset),
    tileset ? tileset : "");
  gtk_entry_set_text(GTK_ENTRY(display_option_dialog_tileset_path),
    tileset_path ? tileset_path : "");
  for ( i = 0; i < 3; i++ ) {
    gtk_entry_set_text(GTK_ENTRY(display_option_dialog_playernames[i]),
			 robot_names[i]);
  }
}

/* apply the selected options */
static void display_option_dialog_apply(void) {
  int i; unsigned int newdp;
  char *newt;
  int restart = 0; /* set to 1 if we need to recreate the display */
  
  for (i=DialogsCentral; i <= DialogsPopup; i++) {
    if ( GTK_TOGGLE_BUTTON(display_option_dialog_dialogposn[i])->active ) 
      newdp = i;
  }
  if ( newdp != dialogs_position ) restart = 1;
  dialogs_position = newdp;
  set_animation(GTK_TOGGLE_BUTTON(display_option_dialog_animation)->active);
  tiletips = GTK_TOGGLE_BUTTON(display_option_dialog_tiletips)->active;
#ifndef WIN32
  iconify_dialogs_with_main = GTK_TOGGLE_BUTTON(display_option_dialog_iconify)->active;
#endif
  i = 
    (GTK_TOGGLE_BUTTON(display_option_dialog_showwall[0])->active ? 0 :
     GTK_TOGGLE_BUTTON(display_option_dialog_showwall[1])->active ? 1 : -1 );
  if ( pref_showwall != i ) {
    restart = 1;
    showwall = pref_showwall = i;
  }
  newt = gtk_entry_get_text(GTK_ENTRY(display_option_dialog_tileset));
  if ( ! tileset ) tileset = "";
  if ( strcmp(newt,tileset) != 0 ) {
    restart = 1;
    tileset = newt;
  }
  newt = gtk_entry_get_text(GTK_ENTRY(display_option_dialog_tileset_path));
  if ( ! tileset_path ) tileset_path = "";
  if ( strcmp(newt,tileset_path) != 0 ) {
    restart = 1;
    tileset_path = newt;
  }
  for ( i = 0; i < 3; i++ ) {
    strcpy(robot_names[i],gtk_entry_get_text(GTK_ENTRY(display_option_dialog_playernames[i])));
    gtk_entry_set_text(GTK_ENTRY(openplayernames[i]),robot_names[i]);
  }
  gtk_widget_hide(display_option_dialog);
  if ( restart ) {
    destroy_display();
    create_display();
  }
}

/* save options */
static void display_option_dialog_save(void) {
  /* first apply */
  display_option_dialog_apply();
  /* then save */
  if ( read_or_update_rcfile(NULL,2,0,0,0) == 0 ) {
    error_dialog_popup("Error updating rc file");
  }
}

/* callback used below */
static void enabler_callback(GtkWidget *w UNUSED, gpointer data)
{
  GameOptionEntry *goe = data;
  goe->enabled = ! goe->enabled;
  make_or_refresh_option_updater(goe,0);
}

/* ditto */
static void make_sensitive(GtkWidget *w)
{
  gtk_widget_set_sensitive(w,1);
}

/* given a GameOptionEntry, create
   or refresh an updater widget, stored in the userdata
   field of the entry.
   Second arg says if this is preference panel (or option panel)
*/
GtkWidget *make_or_refresh_option_updater(GameOptionEntry *goe,int prefsp)
{
  /*      description of option 
          current value   new value  update
  */
  GtkWidget *vbox, *hbox, *ohbox, *w1, *w2,
    *old, *lim, *halflim, *nolim, *resetter;
  GtkAdjustment *adj;
  int dbls, pts;
  int centilims;

  old = goe->userdata;
  if ( ! old ) {
    resetter = gtk_button_new_with_label("Reset"); /* needed early */
  } else {
    resetter = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(old),"reset");
  }
  if ( ! old ) {
    vbox = gtk_vbox_new(0,5);
    gtk_widget_show(vbox);
    if ( prefsp ) {
      w2 = gtk_hbox_new(0,5);
      gtk_widget_show(w2);
      gtk_box_pack_start(GTK_BOX(vbox),w2,0,0,0);
      w1 = gtk_button_new_with_label("Add pref");
      GTK_WIDGET_UNSET_FLAGS(w1,GTK_CAN_FOCUS);
      gtk_object_set_data(GTK_OBJECT(vbox),"enabled",(gpointer) w1);
      gtk_signal_connect(GTK_OBJECT(w1),"clicked",
			 GTK_SIGNAL_FUNC(enabler_callback),(gpointer) goe);
      gtk_widget_show(w1);
      gtk_box_pack_start(GTK_BOX(w2),w1,0,0,0);
      w1 = gtk_label_new(goe->desc);
      gtk_widget_show(w1);
      gtk_box_pack_start(GTK_BOX(w2),w1,0,0,0);
    } else {
      w1 = gtk_label_new(goe->desc);
      gtk_widget_show(w1);
      gtk_box_pack_start(GTK_BOX(vbox),w1,0,0,0);
    }
    if ( prefsp ) {
      ohbox = gtk_hbox_new(0,0);
      gtk_widget_show(ohbox);
      gtk_box_pack_start(GTK_BOX(vbox),ohbox,0,0,0);
      hbox = gtk_hbox_new(0,0);
      gtk_object_set_data(GTK_OBJECT(vbox),"hbox",(gpointer) hbox);
      gtk_widget_show(hbox);
      gtk_box_pack_start(GTK_BOX(ohbox),hbox,1,1,0);
    } else {
      hbox = gtk_hbox_new(0,0);
      gtk_widget_show(hbox);
      gtk_box_pack_start(GTK_BOX(vbox),hbox,0,0,0);
    }
  } else {
    vbox = old;
  }
  switch ( goe->type ) {
  case GOTBool:
    if ( ! old ) {
      w1 = gtk_hbox_new(0,0);
      gtk_widget_show(w1);
      gtk_box_pack_start(GTK_BOX(hbox),w1,0,0,5);
      w2 = gtk_check_button_new_with_label("on/off");
      gtk_signal_connect_object(GTK_OBJECT(w2),"toggled",
				make_sensitive,GTK_OBJECT(resetter));
      gtk_widget_show(w2);
      gtk_box_pack_start(GTK_BOX(w1),w2,0,0,5);
      gtk_object_set_data(GTK_OBJECT(vbox),"checkbox",w2);
    } else {
      w2 = gtk_object_get_data(GTK_OBJECT(vbox),"checkbox");
    }
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w2),
				 goe->value.optbool);
    break;
  case GOTNat:
  case GOTInt:
    if ( ! old ) {
      w1 = gtk_hbox_new(0,0);
      gtk_widget_show(w1);
      gtk_box_pack_start(GTK_BOX(hbox),w1,0,0,5);
      adj = (GtkAdjustment *)gtk_adjustment_new(goe->value.optint,
						(goe->type == GOTNat ) ? 0 : -1000000,1000000,
						1,10,0);
      w2 = gtk_spin_button_new(adj,1.0,0);
      gtk_widget_set_usize(w2,70,0);
      gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(w2),1);
      gtk_signal_connect_object(GTK_OBJECT(w2),"changed",
				make_sensitive,GTK_OBJECT(resetter));
      gtk_widget_show(w2);
      gtk_box_pack_start(GTK_BOX(w1),w2,0,0,5);
      gtk_object_set_data(GTK_OBJECT(vbox),"spinbutton",w2);
    } else {
      w2 = gtk_object_get_data(GTK_OBJECT(vbox),"spinbutton");
      gtk_spin_button_set_value(GTK_SPIN_BUTTON(w2),goe->value.optint);
    }
    break;
  case GOTString:
    if ( ! old ) {
      w1 = gtk_hbox_new(0,0);
      gtk_widget_show(w1);
      gtk_box_pack_start(GTK_BOX(hbox),w1,0,0,5);
      w2 = gtk_entry_new();
      gtk_signal_connect_object(GTK_OBJECT(w2),"changed",
				make_sensitive,GTK_OBJECT(resetter));
      gtk_widget_show(w2);
      gtk_box_pack_start(GTK_BOX(w1),w2,0,0,5);
      gtk_object_set_data(GTK_OBJECT(vbox),"entry",w2);
    } else {
      w2 = gtk_object_get_data(GTK_OBJECT(vbox),"entry");
    }
    gtk_entry_set_text(GTK_ENTRY(w2),goe->value.optstring);
    break;
  case GOTScore:
    pts = goe->value.optscore % 10000;
    dbls = (goe->value.optscore / 10000) % 100;
    centilims = goe->value.optscore/1000000;
    if ( ! old ) {
      w1 = gtk_hbox_new(0,0);
      gtk_widget_show(w1);
      gtk_box_pack_start(GTK_BOX(hbox),w1,0,0,5);
      lim = w2 = gtk_radio_button_new_with_label(NULL,"lim");
      gtk_signal_connect_object(GTK_OBJECT(w2),"toggled",
				make_sensitive,GTK_OBJECT(resetter));
      gtk_widget_show(w2);
      gtk_box_pack_start(GTK_BOX(w1),w2,0,0,5);
      gtk_object_set_data(GTK_OBJECT(vbox),"lim",w2);
      halflim = w2 = gtk_radio_button_new_with_label(gtk_radio_button_group (GTK_RADIO_BUTTON(lim)),"1/2 lim");
      gtk_signal_connect_object(GTK_OBJECT(w2),"toggled",
				make_sensitive,GTK_OBJECT(resetter));
      gtk_widget_show(w2);
      gtk_box_pack_start(GTK_BOX(w1),w2,0,0,5);
      gtk_object_set_data(GTK_OBJECT(vbox),"halflim",w2);
      nolim = w2 = gtk_radio_button_new_with_label(gtk_radio_button_group (GTK_RADIO_BUTTON(lim)),"");
      gtk_signal_connect_object(GTK_OBJECT(w2),"toggled",
				make_sensitive,GTK_OBJECT(resetter));
      gtk_widget_show(w2);
      gtk_box_pack_start(GTK_BOX(w1),w2,0,0,5);
      gtk_object_set_data(GTK_OBJECT(vbox),"nolim",w2);
    } else {
      lim = gtk_object_get_data(GTK_OBJECT(vbox),"lim");
      halflim = gtk_object_get_data(GTK_OBJECT(vbox),"halflim");
      nolim = gtk_object_get_data(GTK_OBJECT(vbox),"nolim");
    }
    switch ( centilims ) {
    case 100:
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(lim),1);
      break;
    case 50:
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(halflim),1);
      break;
    case 0:
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(nolim),1);
      break;
    default:
      warn("Unexpected fraction (%d) of a limit in score; setting to limit",
	   centilims);
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(lim),1);
      break;
    }

    if ( ! old ) {
      adj = (GtkAdjustment *)gtk_adjustment_new(dbls,
						0,100,
						1,10,0);
      w2 = gtk_spin_button_new(adj,1.0,0);
      gtk_widget_set_usize(w2,40,0);
      gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(w2),1);
      gtk_signal_connect_object(GTK_OBJECT(w2),"changed",
				make_sensitive,GTK_OBJECT(resetter));
      gtk_widget_show(w2);
      gtk_box_pack_start(GTK_BOX(w1),w2,0,0,5);
      gtk_object_set_data(GTK_OBJECT(vbox),"dbls",w2);
      w2 = gtk_label_new("dbls");
      gtk_widget_show(w2);
      gtk_box_pack_start(GTK_BOX(w1),w2,0,0,0);
    } else {
      w2 = gtk_object_get_data(GTK_OBJECT(vbox),"dbls");
      gtk_spin_button_set_value(GTK_SPIN_BUTTON(w2),dbls);
    }

    if ( ! old ) {
      adj = (GtkAdjustment *)gtk_adjustment_new(pts,
						0,10000,
						1,10,0);
      w2 = gtk_spin_button_new(adj,1.0,0);
      gtk_widget_set_usize(w2,60,0);
      gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(w2),1);
      gtk_signal_connect_object(GTK_OBJECT(w2),"changed",
				make_sensitive,GTK_OBJECT(resetter));
      gtk_widget_show(w2);
      gtk_box_pack_start(GTK_BOX(w1),w2,0,0,5);
      gtk_object_set_data(GTK_OBJECT(vbox),"pts",w2);
      w2 = gtk_label_new("pts");
      gtk_widget_show(w2);
      gtk_box_pack_start(GTK_BOX(w1),w2,0,0,0);
    } else {
      w2 = gtk_object_get_data(GTK_OBJECT(vbox),"pts");
      gtk_spin_button_set_value(GTK_SPIN_BUTTON(w2),pts);
    }
    break;
  }
  if ( ! old ) {
    /* The button is created at the top so we can feed it to callbacks *
       w1 = gtk_button_new_with_label("Reset");
    */
    w1 = resetter; 
    gtk_widget_show(w1);
    gtk_box_pack_end(GTK_BOX(hbox),w1,0,0,5);
    gtk_object_set_data(GTK_OBJECT(vbox),"reset",w1);
    gtk_signal_connect(GTK_OBJECT(w1),"clicked",
		       GTK_SIGNAL_FUNC(option_reset_callback),
		       (gpointer) goe);
  }
  if ( ! old ) goe->userdata = vbox;
  /* If there's an enabled button, deal with it */
  w1 = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(vbox),"enabled");
  if ( w1 ) {
    hbox = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(vbox),"hbox");
    if ( goe->enabled ) {
      gtk_widget_show(hbox);
      gtk_label_set_text(GTK_LABEL(GTK_BIN(w1)->child),"Remove pref");
    } else { 
      gtk_widget_hide(hbox);
      gtk_label_set_text(GTK_LABEL(GTK_BIN(w1)->child),"Add pref");
    }
  }
  /* make the reset button insensitive */
  gtk_widget_set_sensitive(resetter,0);
  return vbox;
}

/* given an optiontable, pack a list of updating widgets into
   a vbox; or just refresh if already there (2nd arg)
   If third arg is 1, this is prefs panel, else options panel
 */
static GtkWidget *build_or_refresh_optionprefs_panel(GameOptionTable *got, GtkWidget *panel, int prefsp)
{
  int i;
  GtkWidget *vbox,*hs,*u;
  int first = 1;
  
  if ( ! got ) {
    /* destroy the panel's children, if any */
    if ( panel ) {
      gtk_container_foreach(GTK_CONTAINER(panel),(GtkCallback) gtk_widget_destroy,NULL);
    }
  }

  if ( ! panel ) {
    vbox = gtk_vbox_new(0,0);
    gtk_widget_show(vbox);
  } else {
    vbox = panel;
  }
  for ( i = 0 ; got && i < got->numoptions; i++ ) {
    /* if this is actually a filler unknown option, skip it */
    if ( got->options[i].name[0] == 0 ) continue;
    /* likewise for end */
    if ( got->options[i].option == GOEnd ) continue;
    /* if this is the options panel, and the option is not enabled, skip */
    if ( ! prefsp && ! got->options[i].enabled ) continue;
    u = got->options[i].userdata;
    if ( ! first && u == NULL ) {
      hs = gtk_hseparator_new();
      gtk_widget_show(hs);
      gtk_box_pack_start(GTK_BOX(vbox),hs,0,0,0);
    } else first = 0;
    if ( u ) {
      make_or_refresh_option_updater(&got->options[i],prefsp);
    } else {
      u = make_or_refresh_option_updater(&got->options[i],prefsp);
      gtk_box_pack_start(GTK_BOX(vbox),u,0,0,5);
    }
  }
  return vbox;
}

GtkWidget *build_or_refresh_option_panel(GameOptionTable *got, GtkWidget *panel)
{
  return build_or_refresh_optionprefs_panel(got,panel,0);
}

GtkWidget *build_or_refresh_prefs_panel(GameOptionTable *got, GtkWidget *panel)
{
  return build_or_refresh_optionprefs_panel(got,panel,1);
}

void option_reset_callback(GtkWidget *w UNUSED, gpointer data)
{
  make_or_refresh_option_updater((GameOptionEntry *)data,0);
}

void option_updater_callback(GtkWidget *w UNUSED, gpointer data UNUSED)
{
  GameOptionTable *got;
  PMsgSetGameOptionMsg psgom;
  char buf[1024];
  int centilims,dbls,pts;
  int i, changed, val;
  
  if ( ! the_game ) return;
  got = &the_game->option_table;
  for ( i = 0 ; i < got->numoptions ; i++ ) {
    w = got->options[i].userdata;
    if ( ! w ) continue;
    changed = 0;
    psgom.type = PMsgSetGameOption;
    strncpy(psgom.optname,got->options[i].name,16);
    switch ( got->options[i].type ) {
    case GOTBool:
      val = GTK_TOGGLE_BUTTON(gtk_object_get_data(GTK_OBJECT(w),"checkbox"))->active;
      psgom.optvalue = val ? "1" : "0";
      if ( val != got->options[i].value.optbool ) changed = 1;
      break;
    case GOTNat:
    case GOTInt:
      gtk_spin_button_update(GTK_SPIN_BUTTON(gtk_object_get_data(GTK_OBJECT(w),"spinbutton")));
      val = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(gtk_object_get_data(GTK_OBJECT(w),"spinbutton")));
      sprintf(buf,"%d",val);
      psgom.optvalue = buf;
      if ( val != got->options[i].value.optint ) changed = 1;
      break;
    case GOTString:
      psgom.optvalue =
	gtk_entry_get_text(GTK_ENTRY(gtk_object_get_data(GTK_OBJECT(w),"entry")));
      if ( strcmp(psgom.optvalue,got->options[i].value.optstring) != 0 ) changed = 1;
      break;
    case GOTScore:
      centilims = 100*GTK_TOGGLE_BUTTON(gtk_object_get_data(GTK_OBJECT(w),"lim"))->active;
      centilims += 50 * GTK_TOGGLE_BUTTON(gtk_object_get_data(GTK_OBJECT(w),"halflim"))->active;
      gtk_spin_button_update(GTK_SPIN_BUTTON(gtk_object_get_data(GTK_OBJECT(w),"dbls")));
      dbls = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(gtk_object_get_data(GTK_OBJECT(w),"dbls")));
      gtk_spin_button_update(GTK_SPIN_BUTTON(gtk_object_get_data(GTK_OBJECT(w),"pts")));
      pts = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(gtk_object_get_data(GTK_OBJECT(w),"pts")));
      val = 1000000*centilims + 10000 * dbls + pts;
      sprintf(buf,"%d",val);
      psgom.optvalue = buf;
      if ( val != got->options[i].value.optscore ) changed = 1;
      break;
    }
    if ( changed ) send_packet(&psgom);
  }
  gtk_widget_hide(game_option_dialog);
}

void prefs_updater_callback(GtkWidget *w UNUSED, gpointer data UNUSED)
{
  GameOptionTable *got;
  char buf[1024];
  int centilims,dbls,pts;
  int i, changed, val;
  char *vals;
  
  got = &prefs_table;
  for ( i = 0 ; i < got->numoptions ; i++ ) {
    w = got->options[i].userdata;
    if ( ! w ) continue;
    changed = 0;
    switch ( got->options[i].type ) {
    case GOTBool:
      val = GTK_TOGGLE_BUTTON(gtk_object_get_data(GTK_OBJECT(w),"checkbox"))->active;
      if ( val != got->options[i].value.optbool ) changed = 1;
      got->options[i].value.optbool = val;
      break;
    case GOTNat:
    case GOTInt:
      gtk_spin_button_update(GTK_SPIN_BUTTON(gtk_object_get_data(GTK_OBJECT(w),"spinbutton")));
      val = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(gtk_object_get_data(GTK_OBJECT(w),"spinbutton")));
      sprintf(buf,"%d",val);
      if ( val != got->options[i].value.optint ) changed = 1;
      got->options[i].value.optint = val;
      break;
    case GOTString:
      vals =
	gtk_entry_get_text(GTK_ENTRY(gtk_object_get_data(GTK_OBJECT(w),"entry")));
      if ( strcmp(vals,got->options[i].value.optstring) != 0 ) changed = 1;
      strmcpy(got->options[i].value.optstring,vals,128);
      break;
    case GOTScore:
      centilims = 100*GTK_TOGGLE_BUTTON(gtk_object_get_data(GTK_OBJECT(w),"lim"))->active;
      centilims += 50*GTK_TOGGLE_BUTTON(gtk_object_get_data(GTK_OBJECT(w),"halflim"))->active;
      gtk_spin_button_update(GTK_SPIN_BUTTON(gtk_object_get_data(GTK_OBJECT(w),"dbls")));
      dbls = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(gtk_object_get_data(GTK_OBJECT(w),"dbls")));
      gtk_spin_button_update(GTK_SPIN_BUTTON(gtk_object_get_data(GTK_OBJECT(w),"pts")));
      pts = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(gtk_object_get_data(GTK_OBJECT(w),"pts")));
      val = 1000000*centilims + 10000 * dbls + pts;
      if ( val != got->options[i].value.optscore ) changed = 1;
      got->options[i].value.optscore = val;
      break;
    }
    if ( changed ) make_or_refresh_option_updater(&got->options[i],1);
  }
  /* now actually save it */
  read_or_update_rcfile(NULL,0,0,2,0);
  /* and close */
  gtk_widget_hide(game_prefs_dialog);
}

static void apply_game_prefs_callback(GtkWidget *w UNUSED, gpointer data UNUSED)
{
  apply_game_prefs();
  gtk_widget_hide(game_prefs_dialog);
}


void game_option_init(void)
{
  GtkWidget *sbar, *obox, *bbox, *tmp;

  game_option_dialog = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_signal_connect (GTK_OBJECT (game_option_dialog), "delete_event",
                               GTK_SIGNAL_FUNC (gtk_widget_hide), NULL);
  /* must allow shrinking */
  gtk_window_set_policy(GTK_WINDOW(game_option_dialog),TRUE,TRUE,FALSE);
  gtk_window_set_title(GTK_WINDOW(game_option_dialog),"Current Game Options");
  /* reasonable size is ... */
  gtk_window_set_default_size(GTK_WINDOW(game_option_dialog),450,400);

  obox = gtk_vbox_new(0,dialog_vert_spacing);
  gtk_container_set_border_width(GTK_CONTAINER(obox),dialog_border_width);
  gtk_widget_show(obox);
  gtk_container_add(GTK_CONTAINER(game_option_dialog),obox);

  game_option_panel = build_or_refresh_option_panel(the_game ? 
						     &the_game->option_table : NULL,
						     game_option_panel);
  sbar = gtk_scrolled_window_new(NULL,NULL);
  gtk_widget_show(sbar);
  gtk_box_pack_start(GTK_BOX(obox),sbar,1,1,0);
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(sbar),
					game_option_panel);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sbar),
				 GTK_POLICY_NEVER,GTK_POLICY_AUTOMATIC);

  bbox = gtk_hbox_new(1,dialog_button_spacing);
  gtk_widget_show(bbox);
  gtk_box_pack_end(GTK_BOX(obox),bbox,0,0,0);

  tmp = gtk_button_new_with_label("Close");
  GTK_WIDGET_UNSET_FLAGS(tmp,GTK_CAN_FOCUS); /* entry widget shd focus */
  gtk_signal_connect_object(GTK_OBJECT(tmp),"clicked",
			    close_saving_posn,GTK_OBJECT(game_option_dialog));
  gtk_widget_show(tmp);
  gtk_box_pack_end(GTK_BOX(bbox),tmp,1,1,0);

  game_option_prefs_button = tmp = gtk_button_new_with_label("Apply preferences");
  GTK_WIDGET_UNSET_FLAGS(tmp,GTK_CAN_FOCUS); /* entry widget shd focus */
  gtk_signal_connect(GTK_OBJECT(tmp),"clicked",
		     apply_game_prefs_callback,NULL);
  gtk_widget_show(tmp);
  gtk_box_pack_end(GTK_BOX(bbox),tmp,1,1,0);

  game_option_apply_button = tmp = gtk_button_new_with_label("Apply changes");
  GTK_WIDGET_UNSET_FLAGS(tmp,GTK_CAN_FOCUS); /* entry widget shd focus */
  gtk_signal_connect(GTK_OBJECT(tmp),"clicked",
		     option_updater_callback,NULL);
  gtk_widget_show(tmp);
  gtk_box_pack_end(GTK_BOX(bbox),tmp,1,1,0);
}

void game_prefs_init(void)
{
  GtkWidget *sbar, *obox, *bbox, *tmp;

  game_prefs_dialog = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_signal_connect (GTK_OBJECT (game_prefs_dialog), "delete_event",
                               GTK_SIGNAL_FUNC (gtk_widget_hide), NULL);
  /* must allow shrinking */
  gtk_window_set_policy(GTK_WINDOW(game_prefs_dialog),TRUE,TRUE,FALSE);
  gtk_window_set_title(GTK_WINDOW(game_prefs_dialog),"Game Preferences");
  /* reasonable size is ... */
  gtk_window_set_default_size(GTK_WINDOW(game_prefs_dialog),450,400);

  obox = gtk_vbox_new(0,dialog_vert_spacing);
  gtk_container_set_border_width(GTK_CONTAINER(obox),dialog_border_width);
  gtk_widget_show(obox);
  gtk_container_add(GTK_CONTAINER(game_prefs_dialog),obox);

  game_prefs_panel = build_or_refresh_prefs_panel(&prefs_table,
						     game_prefs_panel);
  sbar = gtk_scrolled_window_new(NULL,NULL);
  gtk_widget_show(sbar);
  gtk_box_pack_start(GTK_BOX(obox),sbar,1,1,0);
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(sbar),
					game_prefs_panel);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sbar),
				 GTK_POLICY_NEVER,GTK_POLICY_AUTOMATIC);

  bbox = gtk_hbox_new(1,dialog_button_spacing);
  gtk_widget_show(bbox);
  gtk_box_pack_end(GTK_BOX(obox),bbox,0,0,0);

  tmp = gtk_button_new_with_label("Cancel");
  GTK_WIDGET_UNSET_FLAGS(tmp,GTK_CAN_FOCUS); /* entry widget shd focus */
  gtk_signal_connect_object(GTK_OBJECT(tmp),"clicked",
			    close_saving_posn,GTK_OBJECT(game_prefs_dialog));
  gtk_widget_show(tmp);
  gtk_box_pack_end(GTK_BOX(bbox),tmp,1,1,0);

  tmp = gtk_button_new_with_label("Save changes");
  GTK_WIDGET_UNSET_FLAGS(tmp,GTK_CAN_FOCUS); /* entry widget shd focus */
  gtk_signal_connect(GTK_OBJECT(tmp),"clicked",
		     prefs_updater_callback,NULL);
  gtk_widget_show(tmp);
  gtk_box_pack_end(GTK_BOX(bbox),tmp,1,1,0);
}

/* set the dialog position, changing if necessary */
void set_dialog_posn(DialogPosition p) {
  if ( p != dialogs_position ) {
    /* if the new position is not below, we need to allow the top
       window to shrink down */
    if ( p != DialogsBelow ) {
      gtk_window_set_policy(GTK_WINDOW(topwindow),1,1,1);
    } else {
      /* don't let it shrink */
      gtk_window_set_policy(GTK_WINDOW(topwindow),0,1,0);
    }
    dialogs_position = p;
    create_dialogs();
  }
}

void set_animation(int a) {
  animate = a;
  /* request appropriate pause time */
  if ( the_game ) {
    PMsgSetPlayerOptionMsg spom;
      spom.type = PMsgSetPlayerOption;
    spom.option = PODelayTime;
    spom.ack = 0;
    spom.value = animate ? 10 : 5; /* 0.5 or 1 second min delay */
    spom.text = NULL;
    send_packet(&spom);
  }
}
/* do_chow: if there's only one possible chow, do it, otherwise
   put up a dialog box to choose. Sneakily, this is defined as
   a callback, so that the choosing buttons can call this specifying
   the chow pos, and the main program can call it with AnyPos.
*/

void do_chow(GtkWidget *widg UNUSED, gpointer data) {
  ChowPosition cpos = (ChowPosition)data;
  PMsgChowMsg cm;
  int i,n,j;
  static int lastdiscard;
  TileSet ts;

  cm.type = PMsgChow;
  cm.discard = the_game->serial;

  if ( cpos == AnyPos ) {
    /* we want to avoid working and popping up again if
       for some reason we're already popped up. (So that
       this procedure is idempotent.) */
    if ( GTK_WIDGET_VISIBLE(chow_dialog) && cm.discard == lastdiscard )
      return;
    lastdiscard = cm.discard;
    ts.type = Chow;
    /* set up the boxes */
    for (n=0,i=0,j=0;i<3;i++) {
      if ( player_can_chow(our_player,the_game->tile,i) ) {
	ts.tile = the_game->tile - i;
	tilesetbox_set(&chowtsbs[i],&ts,0);
	tilesetbox_highlight_nth(&chowtsbs[i],i);
	gtk_widget_show(chowbuttons[i]);
	n++; j = i;
      } else {
	gtk_widget_hide(chowtsbs[i].widget);
	gtk_widget_hide(chowbuttons[i]);
      }
    }
    /* if there is only one possibility, don't bother to ask.
       Also if there is no possibility: we'll then get an error
       from the server, which saves us having to worry about it */
    if ( n <= 1 ) {
      cpos = j;
    } else {
      /* pop down the discard dialog */
      gtk_widget_hide(discard_dialog->widget);
      dialog_popup(chow_dialog,DPCentred);
    }
  }

  /* Now we might have found the position */
  if ( cpos != AnyPos ) {
    cm.cpos = cpos;
    send_packet(&cm);
  }
}

/* Generic popup centered over main window.
   Positioning is given by
   DPCentred - centered over main window 
   DPOnDiscard - bottom left corner in same place as discard dialog
   DPErrorPos - for error dialogs: centred over top of main window,
    and offset by a multiple of num_error_dialogs
   DPNone - don't touch the positioning at all '
   DPCentredOnce - centre it on first popup, then don't fiddle '
   DPOnDiscardOnce - on discard dialog first time, then don't fiddle '
   If the widget is not a window, then DPCentredOnce and DPNone
   are equivalent to DPCentred, and DPOnDiscardOnce is equivalent to
   DPOnDiscard.
*/
void dialog_popup(GtkWidget *dialog,DPPosn posn) {
  gint x,y,w,h;
  GtkRequisition r = { 0, 0};

  /* So that we don't work if it's already popped up: */
  /* if ( GTK_WIDGET_VISIBLE(dialog) ) return; */

  if ( ! GTK_IS_WINDOW(dialog) ) {
    if ( posn == DPCentredOnce ) posn = DPCentred;
    if ( posn == DPOnDiscardOnce ) posn = DPOnDiscard;
    if ( posn == DPNone ) posn = DPCentred;
  }

  /* get size of discard widget if necessary */
  if ( (posn == DPOnDiscard || posn == DPOnDiscardOnce) 
       && discard_req.width == 0 ) {
    gtk_widget_size_request(discard_dialog->widget,&discard_req);
    gtk_widget_set_usize(discard_dialog->widget,
			 discard_req.width,
			 discard_req.height);
  }

  gtk_widget_size_request(dialog,&r);

  if ( GTK_IS_WINDOW(dialog) ) {
    gdk_window_get_size(topwindow->window,&w,&h);
    gdk_window_get_deskrelative_origin(topwindow->window,&x,&y);
  } else {
    w = discard_area_alloc.width;
    h = discard_area_alloc.height;
    x = y = 0;
  }

  if ( dialogs_position != DialogsBelow || GTK_IS_WINDOW(dialog) ) {
    if ( posn == DPOnDiscard ) {
      gtk_widget_set_uposition(dialog,
			       x + w/2 - r.width/2
			       - (discard_req.width - r.width)/2,
			       y + h/2 - r.height/2
			       + (discard_req.height - r.height)/2);
    } else if ( posn == DPOnDiscardOnce ) {
      if ( gtk_object_get_data(GTK_OBJECT(dialog),"position-set") ) {
	/* do nothing */
      } else {
	gtk_widget_set_uposition(dialog,
				 x + w/2 - r.width/2
				 - (discard_req.width - r.width)/2,
				 y + h/2 - r.height/2
				 + (discard_req.height - r.height)/2);
	gtk_object_set_data(GTK_OBJECT(dialog),"position-set",(gpointer)1);
      }
    } else if ( posn == DPCentred ) {
      gtk_widget_set_uposition(dialog,
			       x + w/2 - r.width/2,
			       y + h/2 - r.height/2);
    } else if ( posn == DPCentredOnce ) {
      if ( gtk_object_get_data(GTK_OBJECT(dialog),"position-set") ) {
	/* do nothing */
      } else {
	gtk_widget_set_uposition(dialog,
				 x + w/2 - r.width/2,
				 y + h/2 - r.height/2);
	gtk_object_set_data(GTK_OBJECT(dialog),"position-set",(gpointer)1);
      }
    } else if ( posn == DPErrorPos ) {
      gtk_widget_set_uposition(dialog,
			       x + w/2 - r.width/2,
			       y + h/4 - r.height/2 + 10*num_error_dialogs);
    }
  }
  gtk_widget_show(dialog);
}

/* window to display game status. 
     0      1    2...  3....    
   We are WIND; id: n  Name: name
     total score: nnn
   and for right, opposite, left
*/
static char *status_poses[] = { "We are ", "Right is ", "Opp. is ", "Left is " };
static GtkWidget *status_poslabels[4];
static GtkWidget *status_windlabels[4];
static GtkWidget *status_idlabels[4];
static GtkWidget *status_namelabels[4];
static GtkWidget *status_scorelabels[4];
static GtkWidget *status_pwind;
static GtkWidget *status_status;
static GtkWidget *status_tilesleft;

void status_init(void) {
  int i;
  GtkWidget *table, *w;

  status_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_signal_connect(GTK_OBJECT (status_window), "delete_event",
                               GTK_SIGNAL_FUNC (gtk_widget_hide), NULL);
  gtk_window_set_title(GTK_WINDOW(status_window),"Game information");
  gtk_container_set_border_width(GTK_CONTAINER(status_window),
				 dialog_border_width);

  table = gtk_table_new(12,4,0);
  gtk_widget_show(table);
  gtk_container_add(GTK_CONTAINER(status_window),table);
  for ( i = 0 ; i < 4 ; i++ ) {
    status_poslabels[i] = w = gtk_label_new(status_poses[i]);
    gtk_widget_show(w);
    gtk_label_set_justify(GTK_LABEL(w),GTK_JUSTIFY_LEFT);
    gtk_table_attach_defaults(GTK_TABLE(table),w,
		     0,1,2*i,2*i+1);
    status_windlabels[i] = w = gtk_label_new("none");
    gtk_widget_show(w);
    gtk_label_set_justify(GTK_LABEL(w),GTK_JUSTIFY_LEFT);
    gtk_table_attach_defaults(GTK_TABLE(table),w,
		     1,2,2*i,2*i+1);
    status_idlabels[i] = w = gtk_label_new("  ID: 0");
    gtk_widget_show(w);
    gtk_label_set_justify(GTK_LABEL(w),GTK_JUSTIFY_LEFT);
    gtk_table_attach_defaults(GTK_TABLE(table),w,
		     2,3,2*i,2*i+1);
    status_namelabels[i] = w = gtk_label_new("  Name: unknown");
    gtk_widget_show(w);
    gtk_label_set_justify(GTK_LABEL(w),GTK_JUSTIFY_LEFT);
    gtk_table_attach_defaults(GTK_TABLE(table),w,
		     3,4,2*i,2*i+1);
    status_scorelabels[i] = w = gtk_label_new("total score:     0");
    gtk_widget_show(w);
    gtk_label_set_justify(GTK_LABEL(w),GTK_JUSTIFY_LEFT);
    gtk_table_attach_defaults(GTK_TABLE(table),w,
		     1,4,2*i+1,2*i+2);

  }
  status_pwind = w = gtk_label_new("Prevailing wind: none");
  gtk_widget_show(w);
  gtk_table_attach_defaults(GTK_TABLE(table),w,
			    0,4,8,9);

  status_status = w = gtk_label_new("no game");
  gtk_widget_show(w);
  gtk_table_attach_defaults(GTK_TABLE(table),w,
			    0,4,9,10);

  status_tilesleft = w = gtk_label_new("");
  gtk_widget_show(w);
  gtk_table_attach_defaults(GTK_TABLE(table),w,
			    0,4,10,11);

  w = gtk_button_new_with_label("Close");
  gtk_widget_show(w);
  gtk_table_attach_defaults(GTK_TABLE(table),w,
			    0,4,11,12);
  gtk_signal_connect_object(GTK_OBJECT(w),"clicked",
			    close_saving_posn,GTK_OBJECT(status_window));
}

void status_update(int game_over) {
  int i,s;
  char *pn;
  PlayerP p;
  static char buf[256];
  static char *windnames[] = { "none", "East", "South", "West", "North" };

  if ( ! the_game ) return;
  if ( (!game_over) && ! GTK_WIDGET_VISIBLE(status_window) ) return;

  for ( i=0 ; i < 4 ; i++ ) {
    s = (our_seat+i)%NUM_SEATS;
    p = the_game->players[s];
    gtk_label_set_text(GTK_LABEL(status_windlabels[i]),
		       windnames[p->wind]);
    sprintf(buf,"  ID: %d",p->id);
    gtk_label_set_text(GTK_LABEL(status_idlabels[i]),buf);
    sprintf(buf,"  Name: %s",p->name);
    gtk_label_set_text(GTK_LABEL(status_namelabels[i]),buf);
    sprintf(buf,"total score: %5d",p->cumulative_score);
    gtk_label_set_text(GTK_LABEL(status_scorelabels[i]),buf);
  }
  sprintf(buf,"Prevailing wind: %s",windnames[the_game->round]);
  gtk_label_set_text(GTK_LABEL(status_pwind),buf);
  pn = windnames[the_game->player+1]; /* +1 for seat to wind */
  switch (the_game->state) {
  case Dealing:
    sprintf(buf,"Dealing");
    break;
  case DeclaringSpecials:
    sprintf(buf,"%s declaring specials",pn);
    break;
  case Discarding:
    sprintf(buf,"%s to discard",pn);
    break;
  case Discarded:
    sprintf(buf,"%s has discarded",pn);
    break;
  case MahJonging:
    sprintf(buf,"%s has Mah Jong",pn);
    break;
  case HandComplete:
    sprintf(buf,"Hand finished");
    break;
  }
  if ( !the_game->active ) strcat(buf," (Play suspended)");
  if ( game_over ) strcpy(buf,"GAME OVER");
  gtk_label_set_text(GTK_LABEL(status_status),buf);
  sprintf(buf,"%3d tiles left + %2d dead tiles",
	  the_game->wall.live_end-the_game->wall.live_used,
	  the_game->wall.dead_end-the_game->wall.live_end);
  gtk_label_set_text(GTK_LABEL(status_tilesleft),buf);
}

void showraise(GtkWidget *w) {
  gtk_widget_show(w);
  gdk_window_raise(w->window);
  gdk_window_show(w->window);
}

static void about_init(void) {
  GtkWidget *closebutton, *textw, *vbox;
  static char *abouttxt = 
"This is  xmj , part of the Mah-Jong for Unix (etc)\n"
"set of programs.\n"
"Copyright (c) J. C. Bradfield 2000-2001.\n"
"Distributed under the Gnu General Public License.\n"
"This is version " VERSION " (protocol version " STRINGIFY(PROTOCOL_VERSION) ").\n"
"User documentation is in the  xmj  manual page.\n"
"Latest versions, information etc. may be found at\n"
" http://www.stevens-bradfield.com/MahJong/   and\n"
" http://www.dcs.ed.ac.uk/home/jcb/MahJong/   .\n"
"Comments and suggestions should be mailed to\n"
" mahjong@stevens-bradfield.com" ;
  
  about_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_signal_connect (GTK_OBJECT (about_window), "delete_event",
                               GTK_SIGNAL_FUNC (gtk_widget_hide), NULL);
  gtk_window_set_title(GTK_WINDOW(about_window),"About xmj");
  gtk_container_set_border_width(GTK_CONTAINER(about_window),
				 dialog_border_width);

  vbox = gtk_vbox_new(0,dialog_vert_spacing);
  gtk_widget_show(vbox);
  gtk_container_add(GTK_CONTAINER(about_window),vbox);

  textw = gtk_text_new(NULL,NULL);
  gtk_widget_set_usize(textw,300,200);
  gtk_text_insert(GTK_TEXT(textw),NULL,NULL,NULL,abouttxt,-1);
  gtk_widget_show(textw);
  GTK_WIDGET_UNSET_FLAGS(textw,GTK_CAN_FOCUS);
  gtk_box_pack_start(GTK_BOX(vbox),textw,0,0,0);
  
  closebutton = gtk_button_new_with_label("Close");
  gtk_widget_show(closebutton);
  gtk_signal_connect_object(GTK_OBJECT(closebutton),"clicked",
			    close_saving_posn,GTK_OBJECT(about_window));
  gtk_box_pack_start(GTK_BOX(vbox),closebutton,0,0,0);

}

static void nag_callback(GtkWidget *widg UNUSED, gpointer data) {
  nag_state = (int) data;
  read_or_update_rcfile(NULL,0,0,0,2);
  gtk_widget_hide(nag_window);
} 

static void nag_init(void) {
  GtkWidget *yesbutton, *maybebutton, *nobutton, *textw, *vbox;
  char buf[1024];
  static char *nagtxt = ""
"Congratulations: you've completed %d full games using\n"
"this set of programs.\n"
"That suggests that you are getting some enjoyment\n"
"out of them.\n"
"\n"
"If you haven't already, perhaps you would like to\n"
"think about making a small donation to me by way of thanks?\n"
"\n"
"You can do this with a credit card via the home page\n"
"hhtp://www.stevens-bradfield.com/MahJong/";
  
  nag_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_signal_connect (GTK_OBJECT (nag_window), "delete_event",
                               GTK_SIGNAL_FUNC (gtk_widget_hide), NULL);
  gtk_window_set_title(GTK_WINDOW(nag_window),"mu4 juan1");
  gtk_container_set_border_width(GTK_CONTAINER(nag_window),
				 dialog_border_width);

  vbox = gtk_vbox_new(0,dialog_vert_spacing);
  gtk_widget_show(vbox);
  gtk_container_add(GTK_CONTAINER(nag_window),vbox);

  textw = gtk_text_new(NULL,NULL);
  gtk_widget_set_usize(textw,400,200);
  sprintf(buf,nagtxt,completed_games);
  gtk_text_insert(GTK_TEXT(textw),NULL,NULL,NULL,buf,-1);
  gtk_widget_show(textw);
  GTK_WIDGET_UNSET_FLAGS(textw,GTK_CAN_FOCUS);
  gtk_box_pack_start(GTK_BOX(vbox),textw,0,0,0);
  
  yesbutton = gtk_button_new_with_label("Yes, I have/will");
  gtk_widget_show(yesbutton);
  gtk_signal_connect(GTK_OBJECT(yesbutton),"clicked",nag_callback,(gpointer)2);
  gtk_box_pack_start(GTK_BOX(vbox),yesbutton,0,0,0);

  maybebutton = gtk_button_new_with_label("Maybe: remind me again later");
  gtk_widget_show(maybebutton);
  gtk_signal_connect(GTK_OBJECT(maybebutton),"clicked",nag_callback,(gpointer)0);
  gtk_box_pack_start(GTK_BOX(vbox),maybebutton,0,0,0);

  nobutton = gtk_button_new_with_label("No, I won't");
  gtk_widget_show(nobutton);
  gtk_signal_connect(GTK_OBJECT(nobutton),"clicked",nag_callback,(gpointer)1);
  gtk_box_pack_start(GTK_BOX(vbox),nobutton,0,0,0);

}

void nag_popup(void) {
  if ( ! nag_window ) nag_init();
  dialog_popup(nag_window,DPCentred);
}

/* create the dialogs that depend on --dialog-XXX */
void create_dialogs(void) {

  if ( dialogs_position == DialogsBelow ) {
    GtkWidget *t;
    t = gtk_event_box_new(); /* so there's a window to have a style */
    gtk_widget_show(t);
    dialoglowerbox = gtk_hbox_new(0,0);
    gtk_widget_show(dialoglowerbox);
    gtk_container_add(GTK_CONTAINER(t),dialoglowerbox);
    gtk_box_pack_end(GTK_BOX(outerframe),t,0,0,0);
    gtk_widget_show(dialoglowerbox->parent);
  }

  /* create the discard dialog */
  discard_dialog_init();

  /* and the chow dialog */
  chow_dialog_init();

  /* and the declaring specials dialog */
  ds_dialog_init();

  /* and create the turn dialog */
  turn_dialog_init();

  /* now create the scoring dialog */
  scoring_dialog_init();
      
  /* and the continue dialog */
  continue_dialog_init();
}

/* and destroy them */
void destroy_dialogs(void)
{
#define zapit(w) if ( w ) gtk_widget_destroy(w) ; w = NULL ;
  zapit(discard_dialog->widget);
  zapit(chow_dialog);
  zapit(ds_dialog);
  zapit(turn_dialog);
  zapit(scoring_dialog);
  zapit(continue_dialog);
}

static void scoring_showraise(void) { showraise(textwindow); }
static void message_showraise(void) { showraise(messagewindow); }
static void about_showraise(void) { 
  if ( ! about_window ) about_init();
  showraise(about_window); 
}

static void save_showraise(void) {
  if ( ! save_window ) save_init();
  showraise(save_window);
}

void status_showraise(void) { showraise(status_window); status_update(0) ; }

static void display_option_dialog_popup(void) {
  if ( ! display_option_dialog ) display_option_dialog_init();
  display_option_dialog_refresh();
  dialog_popup(display_option_dialog,DPCentredOnce) ; 
}

static void game_option_popup(void) {
  int b;
  if ( ! the_game ) return;
  if ( ! game_option_dialog ) game_option_init();
  game_option_panel = build_or_refresh_option_panel(&the_game->option_table,
						    game_option_panel);
  /* we can't apply options if somebody else is the manager */
  b = (the_game->manager == 0 || the_game->manager == our_id);
  gtk_widget_set_sensitive(game_option_apply_button,b);
  gtk_widget_set_sensitive(game_option_prefs_button,b);
  dialog_popup(game_option_dialog,DPNone);
}

static void game_prefs_popup(void) {
  if ( ! game_prefs_dialog ) game_prefs_init();
  read_or_update_rcfile(NULL,0,0,1,0);
  game_prefs_panel = build_or_refresh_prefs_panel(&prefs_table,game_prefs_panel);
  dialog_popup(game_prefs_dialog,DPNone);
}

static void save_state(void) {
  PMsgSaveStateMsg m;
  m.type = PMsgSaveState;
  m.filename = NULL;
  send_packet(&m);
}

/* the menu items */
static GtkItemFactoryEntry menu_items[] = {
  { "/_Game", NULL, NULL, 0, "<Branch>" },
  { "/Game/_New game...", NULL, open_dialog_popup, 1, NULL },
  { "/Game/_Join game...", NULL, open_dialog_popup, 0, NULL },
  { "/Game/_Resume game...", NULL, open_dialog_popup, 2, NULL },
  { "/Game/_Save", NULL, save_state, 0, NULL },
  { "/Game/Save _as...", NULL, save_showraise, 0, NULL },
  { "/Game/_Close", NULL, close_connection, 0, NULL },
  { "/Game/_Quit", NULL, exit, 0, NULL },
  { "/_Show", NULL, NULL, 0, "<Branch>" },
  { "/Show/_Scoring info", NULL, scoring_showraise, 0, NULL },
  { "/Show/_Game info", NULL, status_showraise, 0, NULL },
  { "/Show/_Messages", NULL, message_showraise, 0, NULL },
  { "/_Options",NULL,NULL,0,"<Branch>" },
  { "/Options/_Display Options...", NULL, display_option_dialog_popup,0,NULL },
  { "/Options/_Current Game Options...", NULL, game_option_popup,0,NULL },
  { "/Options/Game Option _Preferences...", NULL, game_prefs_popup,0,NULL },
  { "/_About", NULL, NULL, 0, "<LastBranch>" },
  { "/About/_About xmj", NULL, about_showraise, 0, NULL },
};
static int nmenu_items = 18;

/* create the menubar, and return it */
GtkWidget *menubar_create(void) {
  GtkItemFactory *item_factory;
  GtkAccelGroup *accel;
  GtkWidget *m;
  int connected = the_game && the_game->fd;

  accel = gtk_accel_group_new();
  item_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR,
				      "<main>",accel);
  gtk_item_factory_create_items(item_factory,nmenu_items,
				menu_items,NULL);
  gtk_window_add_accel_group(GTK_WINDOW(topwindow),accel);

  m = gtk_item_factory_get_widget(item_factory,"<main>");
  gtk_widget_show(m);
  openmenuentry = gtk_item_factory_get_widget(item_factory,
					      "/Game/Join game...");
  assert(openmenuentry);
  gtk_widget_set_sensitive(openmenuentry,!connected);
  newgamemenuentry = gtk_item_factory_get_widget(item_factory,
					      "/Game/New game...");
  assert(newgamemenuentry);
  gtk_widget_set_sensitive(newgamemenuentry,!connected);
  resumegamemenuentry = gtk_item_factory_get_widget(item_factory,
					      "/Game/Resume game...");
  assert(resumegamemenuentry);
  gtk_widget_set_sensitive(resumegamemenuentry,!connected);
  savemenuentry = gtk_item_factory_get_widget(item_factory,
					      "/Game/Save");
  assert(savemenuentry);
  gtk_widget_set_sensitive(savemenuentry,connected);
  saveasmenuentry = gtk_item_factory_get_widget(item_factory,
					      "/Game/Save as...");
  assert(saveasmenuentry);
  gtk_widget_set_sensitive(saveasmenuentry,connected);
  closemenuentry = gtk_item_factory_get_widget(item_factory,"/Game/Close");
  assert(closemenuentry);
  gtk_widget_set_sensitive(closemenuentry,connected);
  gameoptionsmenuentry = gtk_item_factory_get_widget(item_factory,"/Options/Current Game Options...");
  gtk_widget_set_sensitive(gameoptionsmenuentry,connected);
  return m;
}

